from pathlib import Path

import pandas as pd
import pytest

from chickenstats.evolving_hockey.stats import prep_gar, prep_lines, prep_pbp, prep_stats, prep_team, prep_xgar


@pytest.fixture(scope="package")
def raw_pbp():
    filepath = Path("./tests/tests_evolving_hockey/data/raw/raw_pbp.csv")

    raw_pbp = pd.read_csv(filepath, low_memory=False)

    return raw_pbp


@pytest.fixture(scope="package")
def raw_shifts():
    filepath = Path("./tests/tests_evolving_hockey/data/raw/raw_shifts.csv")

    raw_shifts = pd.read_csv(filepath, low_memory=False)

    return raw_shifts


@pytest.fixture(scope="package")
def test_pbp(raw_pbp, raw_shifts):
    pbp = prep_pbp(raw_pbp, raw_shifts)
    return pbp


@pytest.mark.parametrize("columns", ["full", "all", "light"])
def test_prep_pbp(columns, raw_pbp, raw_shifts):
    pbp = prep_pbp(pbp=raw_pbp, shifts=raw_shifts, columns=columns)

    assert isinstance(pbp, pd.DataFrame) is True


def test_prep_pbp_multiple(raw_pbp, raw_shifts):
    raw_pbp = [raw_pbp, raw_pbp]
    raw_shifts = [raw_shifts, raw_shifts]

    pbp = prep_pbp(pbp=raw_pbp, shifts=raw_shifts)

    assert isinstance(pbp, pd.DataFrame) is True


def test_prep_pbp_fail(raw_pbp, raw_shifts):
    raw_pbp = [raw_pbp, raw_pbp]
    raw_shifts = [raw_shifts]

    with pytest.raises(Exception):
        prep_pbp(pbp=raw_pbp, shifts=raw_shifts)


@pytest.mark.parametrize("level", ["game", "period", "season"])
@pytest.mark.parametrize("score", [True, False])
@pytest.mark.parametrize("teammates", [True, False])
@pytest.mark.parametrize("opposition", [True, False])
def test_prep_stats(test_pbp, level, score, teammates, opposition):
    stats = prep_stats(test_pbp, level=level, score=score, teammates=teammates, opposition=opposition)

    assert isinstance(stats, pd.DataFrame) is True


@pytest.mark.parametrize("position", ["f", "d"])
@pytest.mark.parametrize("level", ["game", "period", "season"])
@pytest.mark.parametrize("score", [True, False])
@pytest.mark.parametrize("teammates", [True, False])
@pytest.mark.parametrize("opposition", [True, False])
def test_prep_lines(test_pbp, position, level, score, teammates, opposition):
    lines = prep_lines(
        test_pbp, position=position, level=level, score=score, teammates=teammates, opposition=opposition
    )

    assert isinstance(lines, pd.DataFrame) is True


@pytest.mark.parametrize("level", ["game", "period", "season"])
@pytest.mark.parametrize("strengths", [True, False])
@pytest.mark.parametrize("score", [True, False])
def test_prep_team(test_pbp, level, strengths, score):
    team = prep_team(test_pbp, level, strengths, score)

    assert isinstance(team, pd.DataFrame) is True


@pytest.fixture(scope="package")
def raw_gar_skater():
    filepath = Path("./tests/tests_evolving_hockey/data/raw/raw_gar_skater.csv")

    raw_gar_skater = pd.read_csv(filepath, low_memory=False)

    return raw_gar_skater


@pytest.fixture(scope="package")
def raw_gar_goalie():
    filepath = Path("./tests/tests_evolving_hockey/data/raw/raw_gar_goalie.csv")

    raw_gar_goalie = pd.read_csv(filepath, low_memory=False)

    return raw_gar_goalie


@pytest.fixture(scope="package")
def raw_xgar_skater():
    filepath = Path("./tests/tests_evolving_hockey/data/raw/raw_xgar_skater.csv")

    raw_xgar_skater = pd.read_csv(filepath, low_memory=False)

    return raw_xgar_skater


def test_prep_gar(raw_gar_skater, raw_gar_goalie):
    gar = prep_gar(raw_gar_skater, raw_gar_goalie)

    assert isinstance(gar, pd.DataFrame) is True


def test_prep_xgar(raw_xgar_skater):
    xgar = prep_xgar(raw_xgar_skater)

    assert isinstance(xgar, pd.DataFrame)



import os
from typing import Literal

import numpy as np
import pandas as pd
import polars as pl

from chickenstats.chicken_nhl._validation import LineSchema, PBPSchema, StatSchema, TeamStatSchema
from chickenstats.utilities import ChickenProgress, ChickenProgressIndeterminate

import chickenstats_api


# no cover: start


def _prep_pbp_pandas(pbp: pd.DataFrame) -> list[dict]:
    """Function to prepare a play-by-play dataframe for uploading to the chickenstats API."""
    pbp = pbp.copy()

    goalie_cols = [
        "player_1_api_id",
        "player_2_api_id",
        "player_3_api_id",
        "own_goalie_api_id",
        "opp_goalie_api_id",
        "change_on_goalie_api_id",
        "change_off_goalie_api_id",
        "home_goalie_api_id",
        "away_goalie_api_id",
    ]

    for goalie_col in goalie_cols:
        pbp[goalie_col] = pbp[goalie_col].astype(str).fillna("").astype(str).str.replace(".0", "")

    percent_cols = ["forwards_percent", "opp_forwards_percent"]
    pbp[percent_cols] = pbp[percent_cols].fillna(0.0)

    columns = [x for x in list(PBPSchema.dtypes.keys()) if x in pbp.columns]
    pbp = PBPSchema.validate(pbp[columns])

    pbp = pbp.replace(np.nan, None).replace("nan", None).replace("", None).replace(" ", None)

    api_id_cols = ["player_1_api_id", "player_2_api_id", "player_3_api_id"]
    pbp[api_id_cols] = pbp[api_id_cols].replace("BENCH", None).replace("REFEREE", None).astype("Int64")

    pbp = pbp.to_dict(orient="records")

    return pbp


def _prep_stats_pandas(stats: pd.DataFrame) -> list[dict]:
    """Function to prepare a stats dataframe for uploading to the chickenstats API."""
    stats = stats.copy()

    columns = [x for x in StatSchema.dtypes.keys() if x in stats.columns]

    stats = StatSchema.validate(stats[columns])

    stats = stats.replace(np.nan, None).replace("nan", None)

    stats_id = pd.Series(
        data=(
            stats.game_id.astype(str).copy()
            + "_"
            + "0"
            + stats.period.astype(str).copy()
            + "_"
            + stats.score_state
            + "_"
            + stats.strength_state
            + "_"
            + stats.team
            + "_"
            + stats.api_id.astype(str)
            + "_"
            + stats.forwards_api_id.astype(str).str.replace(", ", "_")
            + "_"
            + stats.defense_api_id.astype(str).str.replace(", ", "_")
            + "_"
            + stats.own_goalie_api_id.astype(str).str.replace(", ", "_")
            + "_"
            + stats.opp_team
            + "_"
            + stats.opp_forwards_api_id.astype(str).str.replace(", ", "_")
            + "_"
            + stats.opp_defense_api_id.astype(str).str.replace(", ", "_")
            + "_"
            + stats.opp_goalie_api_id.astype(str).str.replace(", ", "_")
        ),
        index=stats.index,
        name="id",
        copy=True,
    )

    stats = pd.concat([stats_id, stats], axis=1)

    column_order = [x for x in stats.columns if x != "id"]

    column_order.insert(0, "id")

    stats = stats[column_order]

    stats.id = stats.id.str.replace("_+", "_", regex=True)

    stats = stats.to_dict(orient="records")

    return stats


def _prep_pbp_polars(pbp: pl.DataFrame) -> list[dict]:
    """Function to prepare a play-by-play dataframe for uploading to the chickenstats API."""
    pbp = pbp.with_columns(pl.col(pl.String).replace(old=["", " ", "nan"], new=[None, None, None]))
    pbp = pbp.fill_nan(None)

    pbp = pbp.with_columns(
        forwards_percent=pl.col("forwards_percent").fill_null(0.0),
        opp_forwards_percent=pl.col("opp_forwards_percent").fill_null(0.0),
        player_1_api_id=pl.col("player_1_api_id")
        .replace(old="BENCH", new=None)
        .replace(old="REFEREE", new=None)
        .cast(pl.Int64),
        player_2_api_id=pl.col("player_2_api_id")
        .replace(old="BENCH", new=None)
        .replace(old="REFEREE", new=None)
        .cast(pl.Int64),
        player_3_api_id=pl.col("player_3_api_id")
        .replace(old="BENCH", new=None)
        .replace(old="REFEREE", new=None)
        .cast(pl.Int64),
    )

    pbp = pbp.to_dicts()

    return pbp


def _prep_stats_polars(stats: pl.DataFrame) -> list[dict]:
    """Function to prepare a stats dataframe for uploading to the chickenstats API."""
    stats = stats.with_columns(pl.col(pl.String).replace(old=["nan"], new=[None]))
    stats = stats.fill_nan(None)

    stats = stats.with_columns(
        id=(
            pl.col("game_id").cast(pl.String)
            + "_"
            + "0"
            + pl.col("period").cast(pl.String)
            + "_"
            + pl.col("score_state")
            + "_"
            + pl.col("strength_state")
            + "_"
            + pl.col("team")
            + "_"
            + pl.col("api_id").cast(pl.String)
            + "_"
            + pl.col("forwards_api_id").cast(pl.String).str.replace_all(", ", "_", literal=True).replace(None, "")
            + "_"
            + pl.col("defense_api_id").cast(pl.String).str.replace_all(", ", "_", literal=True).replace(None, "")
            + "_"
            + pl.col("own_goalie_api_id").cast(pl.String).str.replace_all(", ", "_", literal=True).replace(None, "")
            + "_"
            + pl.col("opp_team")
            + "_"
            + pl.col("opp_forwards_api_id").cast(pl.String).str.replace_all(", ", "_", literal=True).replace(None, "")
            + "_"
            + pl.col("opp_defense_api_id").cast(pl.String).str.replace_all(", ", "_", literal=True).replace(None, "")
            + "_"
            + pl.col("opp_goalie_api_id").cast(pl.String).str.replace_all(", ", "_", literal=True).replace(None, "")
        )
    )

    column_order = [x for x in stats.columns if x != "id"]

    column_order.insert(0, "id")

    stats = stats.select(column_order).with_columns(id=pl.col("id").str.replace_all("_+", "_", literal=False))

    stats = stats.to_dicts()

    return stats


class ChickenUser:
    """Generate login tokens and user information for the chickenstats API.

    Parameters:
        username (str):
            The username for the chickenstats API.
            Default is the CHICKENSTATS_USERNAME environment variable
        password (str):
            The password for the chickenstats API.
            Default is the CHICKENSTATS_PASSWORD environment variable
        host (str):
            The URL for the chickenstats API. Default is https://api.chickenstats.com

    Attributes:
        username (str):
            The username given on initialization for the chickenstats API
        password (str):
            The password given on initialization for the chickenstats API
        host (str):
            The URL given on initialization for the chickenstats API
        token (ChickenToken):
            Token object used for logging into the chickenstats API
        access_token (str):
            The bearer token generated after logging into the chickenstats API

    Examples:
        Instantiate the object and generate the user information from default values
        >>> user = ChickenUser()

        You can access the bearer token with the access_token attribute
        >>> access_token = user.access_token

        You can access usernames and passwords from the user object
        >>> username = user.username
        >>> password = user.password

    """

    def __init__(self, username: str | None = None, password: str | None = None, host: str | None = None):
        """Instantiates the user object for the chickenstats API."""
        self.username = username
        self.password = password
        self.host = host

        if not username:
            self.username = os.environ.get("CHICKENSTATS_API_USERNAME")

        if not password:
            self.password = os.environ.get("CHICKENSTATS_API_PASSWORD")

        if not host:
            self.host = os.environ.get("CHICKENSTATS_API_HOST")

            if not self.host:
                self.host = "https://api.chickenstats.com"

        self.configuration = chickenstats_api.Configuration(
            username=self.username, password=self.password, host=self.host
        )

        self.token = None
        self.access_token = None
        self.login()

    def login(self) -> None:
        """Method to log the user into the chickenstats API."""
        with chickenstats_api.ApiClient(self.configuration) as api_client:
            # Create an instance of the API class
            api_instance = chickenstats_api.LoginApi(api_client)

            token = api_instance.login_login_access_token(username=self.username, password=self.password)

            self.token = token
            self.access_token = token.access_token
            self.configuration.access_token = token.access_token

    def reset_password(self, new_password: str) -> None:
        """Method to reset the password for the chickenstats API."""
        self.login()

        with chickenstats_api.ApiClient(self.configuration) as api_client:
            api_instance = chickenstats_api.LoginApi(api_client)
            new_password = chickenstats_api.NewPassword(
                token=self.configuration.access_token, new_password=new_password
            )

            api_instance.login_reset_password(new_password)


class ChickenStats:
    """Generate an API instance for the chickenstats API.

    Parameters:
        username (str):
            The username for the chickenstats API.
            Default is the CHICKENSTATS_USERNAME environment variable
        password (str):
            The password for the chickenstats API.
            Default is the CHICKENSTATS_PASSWORD environment variable
        host (str):
            The URL for the chickenstats API. Default is https://api.chickenstats.com

    Attributes:
        user (ChickenUser):
            A ChickenUser instance for the chickenstats API
        token (dict):
            A dictionary containing the response from the token URL
        access_token (str):
            The bearer token generated after logging into the chickenstats API

    Examples:
        Instantiate the object and generate the user information from default values
        >>> api_instance = ChickenStats()

        You can access the ChickenUser object underlying the instance
        >>> user = api_instance.user
        >>> username = user.username
        >>> user.reset_password(new_password="new_password")

        You can access the bearer token with the access_token attribute
        >>> access_token = api_instance.access_token

        You can then access various API endpoints, starting with play-by-play
        >>> seasons = [2024]
        >>> events = ["GOAL", "SHOT", "MISS"]
        >>> strengths = ["5v5"]
        >>> players = ["FILIP FORSBERG"]
        >>> play_by_play = api_instance.download_pbp(
        ...     season=seasons, event=events, player_1=players, strength_state=strengths
        ... )

        This will download stats data, with the progress bar disabled
        >>> stats = api_instance.download_game_stats(
        ...     season=seasons, player=players, strength_state=strengths, disable_progress_bar=True
        ... )

    """

    def __init__(
        self,
        username: str | None = None,
        password: str | None = None,
        host: str | None = None,
        backend: Literal["polars", "pandas"] = "polars",
    ):
        """Instantiates the ChickenStats object for the chickenstats API."""
        self.user = ChickenUser(username=username, password=password, host=host)
        self.token = self.user.token
        self.access_token = self.user.access_token
        self.backend = backend

    def _finalize_dataframe(self, response) -> pl.DataFrame | pd.DataFrame:
        """Internal method to finalize dataframes when returning stats."""
        if self.backend == "polars":
            df = pl.DataFrame(response)
            df = df.select(col for col in df if col.is_not_null().any())

        if self.backend == "pandas":
            response = [dict(x) for x in response]
            df = pd.DataFrame.from_records(response).dropna(how="all", axis=1)

        return df

    def check_pbp_game_ids(
        self,
        season: list[str | int] | None = None,
        sessions: list[str] | None = None,
        disable_progress_bar: bool = True,
    ) -> list:
        """Check what game IDs are already available from the play-by-play endpoint."""
        with ChickenProgressIndeterminate(disable=disable_progress_bar) as progress:
            pbar_message = "Downloading play-by-play game IDs..."
            progress_task = progress.add_task(pbar_message, total=None, refresh=True)

            progress.start_task(progress_task)
            progress.update(progress_task, total=1, description=pbar_message, refresh=True)

            # Enter a context with an instance of the API client
            with chickenstats_api.ApiClient(self.user.configuration) as api_client:
                # Create an instance of the API class
                api_instance = chickenstats_api.ChickenNhlApi(api_client)

                response = api_instance.chicken_nhl_read_pbp_game_ids(season=season, sessions=sessions)

            progress.update(
                progress_task,
                description="Downloaded play-by-play game IDs",
                completed=True,
                advance=True,
                refresh=True,
            )

        return response

    def check_pbp_play_ids(
        self,
        season: list[str | int] | None = None,
        sessions: list[str] | None = None,
        game_id: list[str | int] | None = None,
        disable_progress_bar: bool = True,
    ) -> list:
        """Check what play IDs are already available from the play-by-play endpoint."""
        with ChickenProgressIndeterminate(disable=disable_progress_bar) as progress:
            pbar_message = "Downloading play-by-play play IDs..."
            progress_task = progress.add_task(pbar_message, total=None, refresh=True)

            progress.start_task(progress_task)
            progress.update(progress_task, total=1, description=pbar_message, refresh=True)

            with chickenstats_api.ApiClient(self.user.configuration) as api_client:
                api_instance = chickenstats_api.ChickenNhlApi(api_client)

                response = api_instance.chicken_nhl_read_pbp_play_ids(season=season, sessions=sessions, game_id=game_id)

            progress.update(progress_task, description=pbar_message, completed=True, advance=True, refresh=True)

        return response

    def upload_pbp(self, pbp: pd.DataFrame | pl.DataFrame, disable_progress_bar: bool = False) -> None:
        """Upload play-by-play data to the chickenstats API. Only available for superusers."""
        with ChickenProgress(disable=disable_progress_bar) as progress:
            pbar_message = "Uploading chicken_nhl play-by-play data..."
            progress_task = progress.add_task(pbar_message, total=None)

            if isinstance(pbp, pd.DataFrame):
                pbp = _prep_pbp_pandas(pbp)

            if isinstance(pbp, pl.DataFrame):
                pbp = _prep_pbp_polars(pbp)

            progress.start_task(progress_task)
            progress_total = len(pbp)
            progress.update(progress_task, total=progress_total, description=pbar_message, refresh=True)

            with chickenstats_api.ApiClient(self.user.configuration) as api_client:
                api_instance = chickenstats_api.PlayByPlayApi(api_client)

                for _idx, row in enumerate(pbp):
                    api_instance.chicken_nhl_create_pbp(row)

                    progress.update(progress_task, description=pbar_message, advance=1, refresh=True)

    def download_pbp(
        self,
        season: list[str | int] | None = None,
        sessions: list[str] | None = None,
        game_id: list[str | int] | None = None,
        event: list[str] | None = None,
        player_1: list[str] | None = None,
        goalie: list[str] | None = None,
        event_team: list[str] | None = None,
        opp_team: list[str] | None = None,
        strength_state: list[str] | None = None,
        disable_progress_bar: bool = False,
    ) -> pl.DataFrame | pd.DataFrame:
        """Download play-by-play data from the chickenstats API.

        Be mindful of your queries, it may fail if the table to return is too large :)

        Parameters:
            season (list[str | int] | None):
                Seasons to download. Defaults to all seasons available
            sessions (list[str] | None):
                Sessions (i.e., regular season or playoffs) to download.
                Defaults to all available.
            game_id (list[str | int] | None):
                Game IDs to download. Defaults to all available.
            event (list[str] | None):
                Events (e.g., GOAL) to download. Defaults to all available.
            player_1 (list[str] | None):
                Event players to download. Defaults to all available.
            goalie (list[str] | None):
                Goalies to download. Defaults to all available.
            event_team (list[str] | None):
                Event teams to download. Defaults to all available.
            opp_team (list[str] | None):
                Opponents to download. Defaults to all available.
            strength_state (list[str] | None):
                Strength states to download. Defaults to all available.
            disable_progress_bar (bool):
                Disables the progress bar if True.

        Examples:
            Download all 5v5 goals for Filip Forsberg in the last five seasons
            >>> cs_instance = ChickenStats()
            >>> forsberg_goals = cs_instance.download_pbp(
            ...     season=[2024, 2023, 2022, 2021, 2020],
            ...     event=["GOAL"],
            ...     player_1=["FILIP FORSBERG"],
            ...     strength_state=["5v5"],
            ... )

            The endpoint is pretty flexible - you can query multiple players and events
            >>> random_shots = cs_instance.download_pbp(
            ...     season=[2024, 2023, 2022, 2021, 2020],
            ...     event=["GOAL", "SHOT", "MISS"],
            ...     player_1=["FILIP FORSBERG", "STEVEN STAMKOS", "MATT DUCHENE"],
            ...     strength_state=["5v5", "4v4", "3v3"],
            ... )

        """
        with ChickenProgressIndeterminate(disable=disable_progress_bar) as progress:
            pbar_message = "Downloading chicken_nhl play-by-play data..."
            progress_task = progress.add_task(pbar_message, total=None, refresh=True)

            progress.start_task(progress_task)
            progress.update(progress_task, total=1, description=pbar_message, refresh=True)

            with chickenstats_api.ApiClient(self.user.configuration) as api_client:
                api_instance = chickenstats_api.PlayByPlayApi(api_client)

                response = api_instance.chicken_nhl_read_pbp(
                    season=season,
                    sessions=sessions,
                    game_id=game_id,
                    event=event,
                    player_1=player_1,
                    goalie=goalie,
                    event_team=event_team,
                    opp_team=opp_team,
                    strength_state=strength_state,
                )

            df = self._finalize_dataframe(response)

            progress.update(
                progress_task,
                description="Downloaded chicken_nhl play-by-play data",
                completed=True,
                advance=True,
                refresh=True,
            )

        return df

    def check_stats_game_ids(
        self,
        season: list[str | int] | None = None,
        sessions: list[str] | None = None,
        disable_progress_bar: bool = True,
    ) -> list:
        """Check what game IDs are already available from the game stats endpoint."""
        with ChickenProgressIndeterminate(disable=disable_progress_bar) as progress:
            pbar_message = "Downloading stats game IDs..."
            progress_task = progress.add_task(pbar_message, total=None, refresh=True)

            progress.start_task(progress_task)
            progress.update(progress_task, total=1, description=pbar_message, refresh=True)

            with chickenstats_api.ApiClient(self.user.configuration) as api_client:
                api_instance = chickenstats_api.StatsApi(api_client)

                response = api_instance.chicken_nhl_read_stats_game_ids(season=season, sessions=sessions)

            progress.update(
                progress_task, description="Downloaded stats game IDs", completed=True, advance=True, refresh=True
            )

        return response

    def upload_stats(self, stats: pd.DataFrame | pl.DataFrame, disable_progress_bar: bool = False) -> None:
        """Upload data for the various stats endpoints. Only available to superusers."""
        with ChickenProgress(disable=disable_progress_bar) as progress:
            pbar_message = "Uploading chicken_nhl stats data..."
            progress_task = progress.add_task(pbar_message, total=None)

            if isinstance(stats, pd.DataFrame):
                stats = _prep_stats_pandas(stats)

            if isinstance(stats, pl.DataFrame):
                stats = _prep_stats_polars(stats)

            progress.start_task(progress_task)
            progress_total = len(stats)
            progress.update(progress_task, total=progress_total, description=pbar_message, refresh=True)

            with chickenstats_api.ApiClient(self.user.configuration) as api_client:
                api_instance = chickenstats_api.StatsApi(api_client)

                for _idx, row in enumerate(stats):
                    api_instance.chicken_nhl_create_stats(row)

                    progress.update(progress_task, description=pbar_message, advance=1, refresh=True)

    def download_game_stats(
        self,
        season: list[str | int] | str | int | None = None,
        sessions: list[str] | str | None = None,
        game_id: list[str | int] | str | int | None = None,
        player: list[str] | str | None = None,
        eh_id: list[str] | str | None = None,
        api_id: list[int] | int | None = None,
        team: list[str] | str | None = None,
        opp_team: list[str] | str | None = None,
        strength_state: list[str] | str | None = None,
        score_state: bool = False,
        teammates: bool = False,
        opposition: bool = False,
        disable_progress_bar: bool = False,
    ) -> pl.DataFrame | pd.DataFrame:
        """Download individual game stats data from the chickenstats API.

        Be mindful of your queries, it may fail if the table to return is too large :)

        Parameters:
            season (list[str | int] | None):
                Seasons to download. Defaults to all seasons available
            sessions (list[str] | None):
                Sessions (i.e., regular season or playoffs) to download.
                Defaults to all available.
            game_id (list[str | int] | None):
                Game IDs to download. Defaults to all available.
            player (list[str] | None):
                Players to download. Defaults to all available.
            eh_id (list[str] | None):
                Evolving Hockey ID for players to download. Defaults to all available.
            api_id (list[int] | int | None):
                API ID for players to download. Defaults to all available.
            team (list[str] | str | None):
                Teams to download. Defaults to all available.
            strength_state (list[str] | None):
                Strength states to download. Defaults to all available.
            disable_progress_bar (bool):
                Disables the progress bar if True.

        Examples:
            Download all 5v5 stats for Filip Forsberg in the last five seasons
            >>> cs_instance = ChickenStats()
            >>> forsberg_stats = cs_instance.download_game_stats(
            ...     season=[2024, 2023, 2022, 2021, 2020], player=["FILIP FORSBERG"], strength_state=["5v5"]
            ... )

            The endpoint is pretty flexible - you can query multiple players
            >>> random_stats = cs_instance.download_game_stats(
            ...     season=[2024, 2023, 2022, 2021, 2020],
            ...     player=["FILIP FORSBERG", "STEVEN STAMKOS", "MATT DUCHENE"],
            ...     strength_state=["5v5", "4v4", "3v3"],
            ... )

        """
        with ChickenProgressIndeterminate(disable=disable_progress_bar) as progress:
            pbar_message = "Downloading chicken_nhl game stats data..."
            progress_task = progress.add_task(pbar_message, total=None, refresh=True)

            progress.start_task(progress_task)
            progress.update(progress_task, total=1, description=pbar_message, refresh=True)

            with chickenstats_api.ApiClient(self.user.configuration) as api_client:
                api_instance = chickenstats_api.StatsApi(api_client)

                response = api_instance.chicken_nhl_read_game_stats(
                    season=season,
                    sessions=sessions,
                    game_id=game_id,
                    player=player,
                    api_id=api_id,
                    eh_id=eh_id,
                    team=team,
                    opp_team=opp_team,
                    strength_state=strength_state,
                    score_state=score_state,
                    teammates=teammates,
                    opposition=opposition,
                )

            df = self._finalize_dataframe(response)

            progress.update(
                progress_task,
                description="Downloaded chicken_nhl game stats data",
                completed=True,
                advance=True,
                refresh=True,
            )

        return df


# no cover: stop



from typing import Literal

import pandas as pd

from chickenstats.chicken_nhl._validation import TeamStatSchema
from chickenstats.evolving_hockey.base import add_positions, munge_pbp, munge_rosters, prep_ind, prep_oi, prep_zones
from chickenstats.evolving_hockey.validation import LineSchema, PBPSchema, StatSchema
from chickenstats.utilities.utilities import ChickenProgress


def prep_pbp(
    pbp: pd.DataFrame | list[pd.DataFrame],
    shifts: pd.DataFrame | list[pd.DataFrame],
    columns: Literal["light", "full", "all"] = "full",
    disable_progress_bar: bool = False,
) -> pd.DataFrame:
    """Prepares a play-by-play dataframe using EvolvingHockey data, but with additional stats and information.

    Columns keyword argument determines information returned. Used in later aggregation
    functions. Returns a DataFrame.

    Parameters:
        pbp (pd.DataFrame):
            Pandas DataFrame of CSV file downloaded from play-by-play query tool at evolving-hockey.com
        shifts (pd.DataFrame):
            Pandas DataFrame of CSV file downloaded from shifts query tool at evolving-hockey.com
        columns (str):
            Whether to return additional columns or more sparse play-by-play dataframe
        disable_progress_bar (bool):
            Whether to disable progress bar

    Returns:
        season (int):
            8-digit season code, e.g., 20192020
        session (str):
            Regular season or playoffs, e.g., R
        game_id (int):
            10-digit game identifier, e.g., 2019020684
        game_date (str):
            Date of game in Eastern time-zone, e.g., 2020-01-09
        event_index (int):
            Unique index number of event, in chronological order, e.g.,
        game_period (int):
            Game period, e.g., 3
        game_seconds (int):
            Game time elapsed in seconds, e.g., 3578
        period_seconds (int):
            Period time elapsed in seconds, e.g., 1178
        clock_time (str):
            Time shown on clock, e.g., 0:22
        strength_state (str):
            Strength state from the perspective of the event team, e.g., 5vE
        score_state (str):
            Score state from the perspective of the event team, e.g., 5v2
        event_type (str):
            Name of the event, e.g., GOAL
        event_description (str):
            Description of the event, e.g., NSH #35 RINNE(1), WRIST, DEF. ZONE, 185 FT.
        event_detail (str | None):
            Additional information about the event, e.g., Wrist
        event_zone (str | None):
            Zone location of event, e.g., DEF
        event_team (str | None):
            3-letter abbreviation of the team for the event, e.g., NSH
        opp_team (str | None):
            3-letter abbreviation of the opposing team for the event, e.g., CHI
        is_home (int | None):
            Dummy variable to signify whether event team is home team, e.g., 0
        coords_x (int | None):
            X coordinates of event, e.g., -96
        coords_y (int | None):
            Y coordinates of event, e.g., 11
        event_player_1 (str):
            Name of the first event player, e.g., PEKKA.RINNE
        event_player_1_id (str):
            Identifier that can be used to match with Evolving Hockey data, e.g., PEKKA.RINNE
        event_player_1_pos (str):
            Player's position for the game, may differ from primary position, e.g., G
        event_player_2 (str):
            Name of the second event player
        event_player_2_id (str):
            Identifier that can be used to match with Evolving Hockey data
        event_player_2_pos (str):
            Player's position for the game, may differ from primary position
        event_player_3 (str):
            Name of the third event player
        event_player_3_id (str):
            Identifier that can be used to match with Evolving Hockey data
        event_player_3_pos (str):
            Player's position for the game, may differ from primary position
        event_length (int):
            Length of time elapsed in seconds since previous event, e.g., 5
        high_danger (int):
            Whether shot event is from high-danger area, e.g., 0
        danger (int):
            Whether shot event is from danger area,
            exclusive of high-danger area, e.g., 0
        pbp_distance (float):
            Distance from opponent net, in feet, according to play-by-play description, e.g., 185.0
        event_distance (float):
            Distance from opponent net, in feet, e.g., 185.326738
        event_angle (float):
            Angle of opponent's net from puck, in degrees, e.g., 57.528808
        opp_strength_state (str):
            Strength state from the perspective of the opposing team, e.g., Ev5
        opp_score_state (str):
            Score state from the perspective of the opposing team, e.g., 2v5
        event_on_f (str):
            Names of the event team's forwards that are on the ice during the event,
            e.g., NICK BONINO, CALLE JARNKROK, MIKAEL GRANLUND
        event_on_f_id (str):
            EH IDs of the event team's forwards that are on the ice during the event,
            e.g., NICK.BONINO, CALLE.JARNKROK, MIKAEL.GRANLUND
        event_on_d (str):
            Names of the event team's defensemen that are on the ice during the event,
            e.g., MATTIAS EKHOLM, ROMAN JOSI
        event_on_d_id (str):
            EH IDs of the event team's defensemen that are on the ice during the event,
            e.g., MATTIAS.EKHOLM, ROMAN.JOSI
        event_on_g (str):
            Name of the goalie for the event team, e.g., PEKKA RINNE
        event_on_g_id (str):
            Identifier for the event team goalie that can be used to match with Evolving Hockey data, e.g., PEKKA.RINNE
        event_on_1 (str):
            Name of one the event team's players that are on the ice during the event,
            e.g., CALLE.JARNKROK
        event_on_1_id (str):
            EH ID of one the event team's players that are on the ice during the event,
            e.g., CALLE.JARNKROK
        event_on_1_pos (str):
            Position of one the event team's players that are on the ice during the event,
            e.g., C
        event_on_2 (str):
            Name of one the event team's players that are on the ice during the event,
            e.g., MATTIAS.EKHOLM
        event_on_2_id (str):
            EH ID of one the event team's players that are on the ice during the event,
            e.g., MATTIAS.EKHOLM
        event_on_2_pos (str):
            Position of one the event team's players that are on the ice during the event,
            e.g., D
        event_on_3 (str):
            Name of one the event team's players that are on the ice during the event,
            e.g., MIKAEL.GRANLUND
        event_on_3_id (str):
            EH ID of one the event team's players that are on the ice during the event,
            e.g., MIKAEL.GRANLUND
        event_on_3_pos (str):
            Position of one the event team's players that are on the ice during the event,
            e.g., C
        event_on_4 (str):
            Name of one the event team's players that are on the ice during the event,
            e.g., NICK.BONINO
        event_on_4_id (str):
            EH ID of one the event team's players that are on the ice during the event,
            e.g., NICK.BONINO
        event_on_4_pos (str):
            Position of one the event team's players that are on the ice during the event,
            e.g., C
        event_on_5 (str):
            Name of one the event team's players that are on the ice during the event,
            e.g., PEKKA.RINNE
        event_on_5_id (str):
            EH ID of one the event team's players that are on the ice during the event,
            e.g., PEKKA.RINNE
        event_on_5_pos (str):
            Position of one the event team's players that are on the ice during the event,
            e.g., G
        event_on_6 (str):
            Name of one the event team's players that are on the ice during the event,
            e.g., ROMAN.JOSI
        event_on_6_id (str):
            EH ID of one the event team's players that are on the ice during the event,
            e.g., ROMAN.JOSI
        event_on_6_pos (str):
            Position of one the event team's players that are on the ice during the event,
            e.g., D
        event_on_7 (str):
            Name of one the event team's players that are on the ice during the event,
            e.g., NaN
        event_on_7_id (str):
            EH ID of one the event team's players that are on the ice during the event,
            e.g., NaN
        event_on_7_pos (str):
            Position of one the event team's players that are on the ice during the event,
            e.g., NaN
        opp_on_f (str):
            Names of the opponent's forwards that are on the ice during the event,
            e.g., ALEX DEBRINCAT, JONATHAN TOEWS, KIRBY DACH, PATRICK KANE
        opp_on_f_id (str):
            EH IDs of the event team's forwards that are on the ice during the event,
            e.g., ALEX.DEBRINCAT, JONATHAN.TOEWS, KIRBY.DACH, PATRICK.KANE
        opp_on_d (str):
            Names of the opposing team's defensemen that are on the ice during the event,
            e.g., DUNCAN KEITH, ERIK GUSTAFSSON
        opp_on_d_id (str):
            EH IDs of the opposing team's defensemen that are on the ice during the event,
            e.g., DUNCAN.KEITH, ERIK.GUSTAFSSON2
        opp_on_g (str):
            Name of the opposing goalie for the event team, e.g., EMPTY NET
        opp_on_g_id (str):
            Identifier for the opposing goalie that can be used to match with Evolving Hockey data, e.g., EMPTY NET
        opp_on_1 (str):
            Name of one the opposing team's players that are on the ice during the event,
            e.g., ALEX.DEBRINCAT
        opp_on_1_id (str):
            EH ID of one the opposing team's players that are on the ice during the event,
            e.g., ALEX.DEBRINCAT
        opp_on_1_pos (str):
            Position of one the opposing team's players that are on the ice during the event,
            e.g., R
        opp_on_2 (str):
            Name of one the opposing team's players that are on the ice during the event,
            e.g., DUNCAN.KEITH
        opp_on_2_id (str):
            EH ID of one the opposing team's players that are on the ice during the event,
            e.g., DUNCAN.KEITH
        opp_on_2_pos (str):
            Position of one the opposing team's players that are on the ice during the event,
            e.g., D
        opp_on_3 (str):
            Name of one the opposing team's players that are on the ice during the event,
            e.g., ERIK.GUSTAFSSON2
        opp_on_3_id (str):
            EH ID of one the opposing team's players that are on the ice during the event,
            e.g., ERIK.GUSTAFSSON2
        opp_on_3_pos (str):
            Position of one the opposing team's players that are on the ice during the event,
            e.g., D
        opp_on_4 (str):
            Name of one the opposing team's players that are on the ice during the event,
            e.g., JONATHAN.TOEWS
        opp_on_4_id (str):
            EH ID of one the opposing team's players that are on the ice during the event,
            e.g., JONATHAN.TOEWS
        opp_on_4_pos (str):
            Position of one the opposing team's players that are on the ice during the event,
            e.g., C
        opp_on_5 (str):
            Name of one the opposing team's players that are on the ice during the event,
            e.g., KIRBY.DACH
        opp_on_5_id (str):
            EH ID of one the opposing team's players that are on the ice during the event,
            e.g., KIRBY.DACH
        opp_on_5_pos (str):
            Position of one the opposing team's players that are on the ice during the event,
            e.g., C
        opp_on_6 (str):
            Name of one the opposing team's players that are on the ice during the event,
            e.g., PATRICK.KANE
        opp_on_6_id (str):
            EH ID of one the opposing team's players that are on the ice during the event,
            e.g., PATRICK.KANE
        opp_on_6_pos (str):
            Position of one the opposing team's players that are on the ice during the event,
            e.g., R
        opp_on_7 (str):
            Name of one the opposing team's players that are on the ice during the event,
            e.g., NaN
        opp_on_7_id (str):
            EH ID of one the opposing team's players that are on the ice during the event,
            e.g., NaN
        opp_on_7_pos (str):
            Position of one the opposing team's players that are on the ice during the event,
            e.g., NaN
        change (int):
            Dummy variable to indicate whether event is a change, e.g., 0
        zone_start (str):
            Zone where the changed, e.g., OFF or OTF
        num_on (int | None):
            Number of players entering the ice, e.g., 6
        num_off (int | None):
            Number of players exiting the ice, e.g., 0
        players_on (str):
            Names of players on, in jersey order,
            e.g., FILIP FORSBERG, ALEX CARRIER, ROMAN JOSI, MIKAEL GRANLUND, JUUSE SAROS, MATT DUCHENE
        players_on_id (str):
            Evolving Hockey IDs of players on, in jersey order,
            e.g., FILIP.FORSBERG, ALEX.CARRIER, ROMAN.JOSI, MIKAEL.GRANLUND, JUUSE.SAROS, MATT.DUCHENE
        players_on_pos (str):
            Positions of players on, in jersey order,
            e.g., L, D, D, C, G, C
        players_off (str):
            Names of players off, in jersey order
        players_off_id (str):
            Evolving Hockey IDs of players off, in jersey order
        players_off_pos (str):
            Positions of players off, in jersey order
        shot (int):
            Dummy variable to indicate whether event is a shot, e.g., 1
        shot_adj (float):
            Score and venue-adjusted shot value, e.g., 0
        goal (int):
            Dummy variable to indicate whether event is a goal, e.g., 1
        goal_adj (float):
            Score and venue-adjusted shot value, e.g., 0
        pred_goal (float):
            Predicted goal value (xG), e.g., 0.482589
        pred_goal_adj (float):
            Score and venue-adjusted predicted goal (xG) value, e.g., 0
        miss (int):
            Dummy variable to indicate whether event is a missed shot, e.g., 0
        block (int):
            Dummy variable to indicate whether event is a block, e.g., 0
        corsi (int):
            Dummy variable to indicate whether event is a corsi event, e.g., 1
        corsi_adj (float):
            Score and venue-adjusted corsi value, e.g., 0
        fenwick (int):
            Dummy variable to indicate whether event is a fenwick event, e.g., 1
        fenwick_adj (float):
             Score and venue-adjusted fenwick value, e.g., 0
        hd_shot (int):
            Dummy variable to indicate whether event is a high-danger shot event, e.g., 0
        hd_goal (int):
            Dummy variable to indicate whether event is a high-danger goal event, e.g., 0
        hd_miss (int):
            Dummy variable to indicate whether event is a high-danger miss event, e.g., 0
        hd_fenwick (int):
            Dummy variable to indicate whether event is a high-danger fenwick event, e.g., 0
        fac (int):
            Dummy variable to indicate whether event is a faceoff, e.g., 0
        hit (int):
            Dummy variable to indicate whether event is a hit, e.g., 0
        give (int):
            Dummy variable to indicate whether event is a giveaway, e.g., 0
        take (int):
            Dummy variable to indicate whether event is a takeaway, e.g., 0
        pen0 (int):
            Dummy variable to indicate whether event is a penalty with no minutes, e.g., 0
        pen2 (int):
            Dummy variable to indicate whether event is a two-minute penalty, e.g., 0
        pen4 (int):
            Dummy variable to indicate whether event is a four-minute penalty, e.g., 0
        pen5 (int):
            Dummy variable to indicate whether event is a five-minute penalty, e.g., 0
        pen10 (int):
            Dummy variable to indicate whether event is a ten-minute penalty, e.g., 0
        stop (int):
            Dummy variable to indicate whether event is a stoppage, e.g., 0
        ozf (int):
            Dummy variable to indicate whether event is an offensive zone faceoff e.g., 0
        nzf (int):
            Dummy variable to indicate whether event is a neutral zone faceoff, e.g., 0
        dzf (int):
            Dummy variable to indicate whether event is a defensive zone faceoff, e.g., 0
        ozs (int):
            Dummy variable to indicate whether an event is an offensive zone change, e.g., 0
        nzs (int):
            Dummy variable to indicate whether an event is a neutral zone change, e.g., 0
        dzs (int):
            Dummy variable to indicate whether an event is an defensive zone change, e.g., 0
        otf (int):
            Dummy variable to indicate whether an event is an on-the-fly change, e.g., 0

    Examples:
        Play-by-play DataFrame
        >>> shifts_raw = pd.read_csv("./raw_shifts.csv")
        >>> pbp_raw = pd.read_csv("./raw_pbp.csv")
        >>> pbp = prep_pbp(pbp_raw, shifts_raw)

    """
    with ChickenProgress(disable=disable_progress_bar) as progress:
        if isinstance(pbp, pd.DataFrame):
            progress_total = 1

            pbp = [pbp]

        elif isinstance(pbp, list):
            progress_total = len(pbp)

        if isinstance(shifts, pd.DataFrame):
            shifts = [shifts]

        if len(pbp) != len(shifts):
            raise Exception("Number of play-by-play and shift CSV files does not match")

        pbar_message = "Prepping play-by-play data..."

        csv_task = progress.add_task(pbar_message, total=progress_total)

        pbp_concat = []

        for idx, (pbp_raw, shifts_raw) in enumerate(zip(pbp, shifts, strict=False)):
            rosters = munge_rosters(shifts_raw)

            pbp_clean = munge_pbp(pbp_raw)

            pbp_clean = add_positions(pbp_clean, rosters)

            cols = [
                "id",
                "season",
                "session",
                "game_id",
                "game_date",
                "event_index",
                "game_period",
                "game_seconds",
                "period_seconds",
                "clock_time",
                "strength_state",
                "score_state",
                "event_type",
                "event_description",
                "event_detail",
                "event_zone",
                "event_team",
                "opp_team",
                "is_home",
                "coords_x",
                "coords_y",
                "event_player_1",
                "event_player_1_id",
                "event_player_1_pos",
                "event_player_2",
                "event_player_2_id",
                "event_player_2_pos",
                "event_player_3",
                "event_player_3_id",
                "event_player_3_pos",
                "event_length",
                "high_danger",
                "danger",
                "pbp_distance",
                "event_distance",
                "event_angle",
                "event_on_f",
                "event_on_f_id",
                "event_on_d",
                "event_on_d_id",
                "event_on_g",
                "event_on_g_id",
                "opp_on_f",
                "opp_on_f_id",
                "opp_on_d",
                "opp_on_d_id",
                "opp_on_g",
                "opp_on_g_id",
                "change",
                "zone_start",
                "num_on",
                "num_off",
                "players_on",
                "players_on_id",
                "players_on_pos",
                "players_off",
                "players_off_id",
                "players_off_pos",
                "shot",
                "shot_adj",
                "goal",
                "goal_adj",
                "pred_goal",
                "pred_goal_adj",
                "miss",
                "block",
                "corsi",
                "corsi_adj",
                "fenwick",
                "fenwick_adj",
                "hd_shot",
                "hd_goal",
                "hd_miss",
                "hd_fenwick",
                "fac",
                "hit",
                "give",
                "take",
                "pen0",
                "pen2",
                "pen4",
                "pen5",
                "pen10",
                "stop",
                "ozf",
                "nzf",
                "dzf",
                "ozs",
                "nzs",
                "dzs",
                "otf",
            ]

            if columns in ["full", "all"]:
                event_cols = [
                    "event_on_1",
                    "event_on_1_id",
                    "event_on_1_pos",
                    "event_on_2",
                    "event_on_2_id",
                    "event_on_2_pos",
                    "event_on_3",
                    "event_on_3_id",
                    "event_on_3_pos",
                    "event_on_4",
                    "event_on_4_id",
                    "event_on_4_pos",
                    "event_on_5",
                    "event_on_5_id",
                    "event_on_5_pos",
                    "event_on_6",
                    "event_on_6_id",
                    "event_on_6_pos",
                    "event_on_7",
                    "event_on_7_id",
                    "event_on_7_pos",
                ]

                event_pos = cols.index("event_on_g_id") + 1

                cols[event_pos:event_pos] = event_cols

                opp_cols = [
                    "opp_on_1",
                    "opp_on_1_id",
                    "opp_on_1_pos",
                    "opp_on_2",
                    "opp_on_2_id",
                    "opp_on_2_pos",
                    "opp_on_3",
                    "opp_on_3_id",
                    "opp_on_3_pos",
                    "opp_on_4",
                    "opp_on_4_id",
                    "opp_on_4_pos",
                    "opp_on_5",
                    "opp_on_5_id",
                    "opp_on_5_pos",
                    "opp_on_6",
                    "opp_on_6_id",
                    "opp_on_6_pos",
                    "opp_on_7",
                    "opp_on_7_id",
                    "opp_on_7_pos",
                ]

                opp_pos = cols.index("opp_on_g_id") + 1

                cols[opp_pos:opp_pos] = opp_cols

                other_cols = ["opp_strength_state", "opp_score_state"]

                other_pos = cols.index("event_angle") + 1

                cols[other_pos:other_pos] = other_cols

            if columns == "all":
                more_cols = [
                    "home_zone",
                    "home_team",
                    "away_team",
                    "home_goalie",
                    "away_goalie",
                    "home_skaters",
                    "away_skaters",
                    "home_score",
                    "away_score",
                    "home_zonestart",
                    "face_index",
                    "pen_index",
                    "shift_index",
                    "game_score_state",
                    "game_strength_state",
                ]

                pos = cols.index("is_home") + 1

                cols[pos:pos] = more_cols

            cols = [x for x in cols if x in pbp_clean]

            pbp_clean = pbp_clean[cols]

            cols = [x for x in list(PBPSchema.dtypes.keys()) if x in pbp_clean.columns]

            pbp_clean = PBPSchema.validate(pbp_clean[cols])

            pbp_concat.append(pbp_clean)

            if progress_total == 1 or idx + 1 == progress_total:
                pbp_clean = pd.concat(pbp_concat, ignore_index=True)

                pbar_message = "Finished loading play-by-play data"

            progress.update(csv_task, description=pbar_message, advance=1, refresh=True)

    return pbp_clean


# Function combining the on-ice and individual stats
def prep_stats(
    pbp: pd.DataFrame,
    level: Literal["period", "game", "session", "season"] = "game",
    score: bool = False,
    teammates: bool = False,
    opposition: bool = False,
    disable_progress_bar: bool = False,
) -> pd.DataFrame:
    """Prepares an individual and on-ice stats dataframe using EvolvingHockey data.

    Aggregates to desired level. Capable of returning cuts that account for strength state,
    period, score state, teammates, and opposition.

    Returns a Pandas DataFrame.

    Parameters:
        pbp (pd.DataFrame):
            Dataframe from the prep_pbp function with the default columns argument
        level (str):
            Level to aggregate stats, e.g., 'game'
        score (bool):
            Whether to aggregate to score state level
        teammates (bool):
            Whether to account for teammates when aggregating
        opposition (bool):
            Whether to account for opposition when aggregating
        disable_progress_bar (bool):
            Whether to disable progress bar

    Returns:
        season (int):
            8-digit season code, e.g., 20232024
        session (str):
            Regular season or playoffs, e.g., R
        game_id (int):
            10-digit game identifier, e.g., 2023020015
        game_date (str):
            Date of game in Eastern time-zone, e.g., 2023-10-12
        player (str):
            Name of the player, e.g., FILIP.FORSBERG
        player_id (str):
            Player EH ID, e.g., FILIP.FORSBERG
        position (str):
            Player's position, e.g., L
        team (str):
            3-letter abbreviation of the player's team, e.g., NSH
        opp_team: object
            3-letter abbreviation of the opposing team, e.g., SEA
        strength_state (str):
            Strength state from the perspective of the event team, e.g., 5v5
        score_state (str):
            Score state from the perspective of the event team, e.g., 0v0
        game_period (int):
            Game period, e.g., 1
        forwards (str):
            Names of the event team's forwards that are on the ice during the event,
            e.g., FILIP.FORSBERG, JUUSO.PARSSINEN, RYAN.O'REILLY
        forwards_id (str):
            EH IDs of the event team's forwards that are on the ice during the event,
            e.g., FILIP.FORSBERG, JUUSO.PARSSINEN, RYAN.O'REILLY
        defense (str):
            Names of the event team's defensemen that are on the ice during the event,
            e.g., ALEX.CARRIER, RYAN.MCDONAGH
        defense_id (str):
            EH IDs of the event team's defensemen that are on the ice during the event,
            e.g., ALEX.CARRIER, RYAN.MCDONAGH
        own_goalie (str):
            Name of the goalie for the event team, e.g., JUUSE.SAROS
        own_goalie_id (str):
            Identifier for the event team goalie that can be used to match with Evolving Hockey data, e.g., JUUSE.SAROS
        opp_forwards (str):
            Names of the opponent's forwards that are on the ice during the event,
            e.g., JARED.MCCANN, JORDAN.EBERLE, MATTY.BENIERS
        opp_forwards_id (str):
            EH IDs of the event team's forwards that are on the ice during the event,
            e.g., JARED.MCCANN, JORDAN.EBERLE, MATTY.BENIERS
        opp_defense(str):
            Names of the opposing team's defensemen that are on the ice during the event,
            e.g., JAMIE.OLEKSIAK, WILLIAM.BORGEN
        opp_defense_id (str):
            EH IDs of the opposing team's defensemen that are on the ice during the event,
            e.g., JAMIE.OLEKSIAK, WILLIAM.BORGEN
        opp_goalie (str):
            Name of the opposing goalie for the event team, e.g., PHILIPP.GRUBAUER
        opp_goalie_id (str):
            Identifier for the opposing goalie that can be used to match with Evolving Hockey data,
            e.g., PHILIPP.GRUBAUER
        toi (float):
            Time on-ice in minutes, e.g., 1.616667
        g (float):
            Number of individual goals scored, e.g, 0
        a1 (float):
            Number of primary assists, e.g, 0
        a2 (float):
            Number of secondary assists, e.g, 0
        isf (float):
            Number of indiviudal shots registered, e.g., 0
        iff (float):
            Number of indiviudal fenwick events registered, e.g., 0
        icf (float):
            Number of indiviudal corsi events registered, e.g., 0
        ixg (float):
            Sum value of individual predicted goals (xG), e.g., 0
        gax (float):
            Sum value of goals scored above expected, e.g., 0
        ihdg (float):
            Sum value of individual high-danger goals scored, e.g., 0
        ihdf (float):
            Sum value of individual high-danger fenwick events registered, e.g., 0
        ihdsf (float):
            Sum value of individual high-danger shots taken, e.g., 0
        ihdm (float):
            Sum value of individual high-danger shots missed, e.g., 0
        imsf (float):
            Sum value of individual missed shots, 0
        isb (float):
            Sum value of shots taken that were ultimately blocked, e.g., 0
        ibs (float):
            Sum value of opponent shots taken that the player ultimately blocked, e.g., 0
        igive (float):
            Sum of individual giveaways, e.g., 0
        itake (float):
            Sum of individual takeaways, e.g., 0
        ihf (float):
            Sum of individual hits for, e.g., 0
        iht (float):
            Sum of individual hits taken, e.g., 0
        ifow (float):
            Sum of individual faceoffs won, e.g., 0
        ifol (float):
            Sum of individual faceoffs lost, e.g., 0
        iozfw (float):
            Sum of individual faceoffs won in offensive zone, e.g., 0
        iozfl (float):
            Sum of individual faceoffs lost in offensive zone, e.g., 0
        inzfw (float):
            Sum of individual faceoffs won in neutral zone, e.g., 0
        inzfl (float):
            Sum of individual faceoffs lost in neutral zone, e.g., 0
        idzfw (float):
            Sum of individual faceoffs won in defensive zone, e.g., 0
        idzfl (float):
            Sum of individual faceoffs lost in defensive zone, e.g., 0
        a1_xg (float):
            Sum of xG from primary assists, e.g., 0
        a2_xg (float):
            Sum of xG from secondary assists, e.g., 0
        ipent0 (float):
            Sum of individual 0-minute penalties taken, e.g., 0
        ipent2 (float):
            Sum of individual 2-minute penalties taken, e.g., 0
        ipent4 (float):
            Sum of individual 4-minute penalties taken, e.g., 0
        ipent5 (float):
            Sum of individual 5-minute penalties taken, e.g., 0
        ipent10 (float):
            Sum of individual 10-minute penalties taken, e.g., 0
        ipend0 (float):
            Sum of individual 0-minute penalties drawn, e.g., 0
        ipend2 (float):
            Sum of individual 2-minute penalties drawn, e.g., 0
        ipend4 (float):
            Sum of individual 4-minute penalties drawn, e.g., 0
        ipend5 (float):
            Sum of individual 5-minute penalties drawn, e.g., 0
        ipend10 (float):
            Sum of individual 10-minute penalties drawn, e.g., 0
        ozs (float):
            Sum of changes with offensive zone starts, e.g., 0
        nzs (float):
            Sum of changes with neutral zone starts, e.g., 0
        dzs (float):
            Sum of changes with defensive zone starts, e.g., 1
        otf (float):
            Sum of changes on-the-fly, e.g., 0
        gf (float):
            Sum of goals scored while player is on-ice, e.g., 0
        gf_adj (float):
            Sum of venue- and score-adjusted goals scored while player is on-ice, e.g., 0
        hdgf (float):
            Sum of high-danger goals scored while player is on-ice, e.g., 0
        ga (float):
            Sum of goals allowed while player is on-ice, e.g., 0
        ga_adj (float):
            Sum of venue- and score-adjusted goals allowed while player is on-ice, e.g., 0
        hdga (float):
            Sum of high-danger goals allowed while player is on-ice, e.g., 0
        xgf (float):
            Sum of expected goals generated while player is on-ice, e.g., 0.017266
        xgf_adj (float):
            Sum of venue- and score-adjusted expected goals generated while player is on-ice, e.g., 0.016472
        xga (float):
            Sum of expected goals allowed while player is on-ice, e.g., 0.123475
        xga_adj (float):
            Sum of venue- and score-adjusted expected goals allowed while player is on-ice, e.g., 0.129772
        sf (float):
            Sum of shots taken while player is on-ice, e.g., 1
        sf_adj (float):
            Sum of venue- and score-adjusted shots taken while player is on-ice, e.g., .972
        hdsf (float):
            Sum of high-danger shots taken while player is on-ice, e.g., 0
        sa (float):
            Sum of shots allowed while player is on-ice, e.g., 0
        sa_adj (float):
            Sum of venue- and score-adjusted shots allowed while player is on-ice, e.g., 0
        hdsa (float):
            Sum of high-danger shots allowed while player is on-ice, e.g., 0
        ff (float):
            Sum of fenwick events generated while player is on-ice, e.g., 1
        ff_adj (float):
            Sum of venue- and score-adjusted fenwick events generated while player is on-ice, e.g., 0.968
        hdff (float):
            Sum of high-danger fenwick events generated while player is on-ice, e.g., 0
        fa (float):
            Sum of fenwick events allowed while player is on-ice, e.g., 1
        fa_adj (float):
            Sum of venue- and score-adjusted fenwick events allowed while player is on-ice, e.g., 1.034
        hdfa (float):
            Sum of high-danger fenwick events allowed while player is on-ice, e.g., 1
        cf (float):
            Sum of corsi events generated while player is on-ice, e.g., 1
        cf_adj (float):
            Sum of venue- and score-adjusted corsi events generated while player is on-ice, e.g., 0.970
        ca (float):
            Sum of corsi events allowed while player is on-ice, e.g., 2
        ca_adj (float):
            Sum of venue- and score-adjusted corsi events allowed while player is on-ice, e.g., 2.064
        bsf (float):
            Sum of shots taken that were ultimately blocked while player is on-ice, e.g., 0
        bsa (float):
            Sum of shots allowed that were ultimately blocked while player is on-ice, e.g., 1
        msf (float):
            Sum of shots taken that missed net while player is on-ice, e.g., 0
        hdmsf (float):
            Sum of high-danger shots taken that missed net while player is on-ice, e.g., 0
        msa (float):
            Sum of shots allowed that missed net while player is on-ice, e.g., 1
        hdmsa (float):
            Sum of high-danger shots allowed that missed net while player is on-ice, e.g., 1
        hf (float):
            Sum of hits dished out while player is on-ice, e.g., 0
        ht (float):
            Sum of hits taken while player is on-ice, e.g., 0
        ozf (float):
            Sum of offensive zone faceoffs that occur while player is on-ice, e.g., 0
        nzf (float):
            Sum of neutral zone faceoffs that occur while player is on-ice, e.g., 0
        dzf (float):
            Sum of defensive zone faceoffs that occur while player is on-ice, e.g., 1
        fow (float):
            Sum of faceoffs won while player is on-ice, e.g., 1
        fol (float):
            Sum of faceoffs lost while player is on-ice, e.g., 0
        ozfw (float):
            Sum of offensive zone faceoffs won while player is on-ice, e.g., 0
        ozfl (float):
            Sum of offensive zone faceoffs lost while player is on-ice, e.g., 1
        nzfw (float):
            Sum of neutral zone faceoffs won while player is on-ice, e.g., 0
        nzfl (float):
            Sum of neutral zone faceoffs lost while player is on-ice, e.g., 0
        dzfw (float):
            Sum of defensive zone faceoffs won while player is on-ice, e.g., 1
        dzfl (float):
            Sum of defensive zone faceoffs lost while player is on-ice, e.g., 0
        pent0 (float):
            Sum of individual 0-minute penalties taken while player is on-ice, e.g., 0
        pent2 (float):
            Sum of individual 2-minute penalties taken while player is on-ice, e.g., 0
        pent4 (float):
            Sum of individual 4-minute penalties taken while player is on-ice, e.g., 0
        pent5 (float):
            Sum of individual 5-minute penalties taken while player is on-ice, e.g., 0
        pent10 (float):
            Sum of individual 10-minute penalties taken while player is on-ice, e.g., 0
        pend0 (float):
            Sum of individual 0-minute penalties drawn while player is on-ice, e.g., 0
        pend2 (float):
            Sum of individual 2-minute penalties drawn while player is on-ice, e.g., 0
        pend4 (float):
            Sum of individual 4-minute penalties drawn while player is on-ice, e.g., 0
        pend5 (float):
            Sum of individual 5-minute penalties drawn while player is on-ice, e.g., 0
        pend10 (float):
            Sum of individual 10-minute penalties drawn while player is on-ice, e.g., 0

    Examples:
        Basic play-by-play DataFrame
        >>> shifts_raw = pd.read_csv("./raw_shifts.csv")
        >>> pbp_raw = pd.read_csv("./raw_pbp.csv")
        >>> pbp = prep_pbp(pbp_raw, shifts_raw)

        Basic game-level stats, with no teammates or opposition
        >>> stats = prep_stats(pbp)

        Period-level stats, grouped by teammates
        >>> stats = prep_stats(pbp, level="period", teammates=True)

        Session-level (e.g., regular seasion) stats, grouped by teammates and opposition
        >>> stats = prep_stats(pbp, level="session", teammates=True, opposition=True)

    """
    with ChickenProgress(disable=disable_progress_bar) as progress:
        pbar_message = "Prepping stats data..."

        stats_task = progress.add_task(pbar_message, total=1)

        ind = prep_ind(pbp, level, score, teammates, opposition)

        oi = prep_oi(pbp, level, score, teammates, opposition)

        zones = prep_zones(pbp, level, score, teammates, opposition)

        merge_cols = [
            "season",
            "session",
            "game_id",
            "game_date",
            "player",
            "player_id",
            "position",
            "team",
            "opp_team",
            "strength_state",
            "score_state",
            "game_period",
            "forwards",
            "forwards_id",
            "defense",
            "defense_id",
            "own_goalie",
            "own_goalie_id",
            "opp_forwards",
            "opp_forwards_id",
            "opp_defense",
            "opp_defense_id",
            "opp_goalie",
            "opp_goalie_id",
        ]

        merge_cols = [x for x in merge_cols if x in ind.columns and x in oi.columns and x in zones.columns]

        stats = oi.merge(ind, how="left", left_on=merge_cols, right_on=merge_cols).fillna(0)

        stats = stats.merge(zones, how="left", left_on=merge_cols, right_on=merge_cols).fillna(0)

        stats = stats.loc[stats.toi > 0].reset_index(drop=True).copy()

        columns = [x for x in StatSchema.dtypes if x in stats.columns]

        stats = StatSchema.validate(stats[columns])

        pbar_message = "Finished prepping stats data"

        progress.update(stats_task, description=pbar_message, advance=1, refresh=True)

    return stats


# Function to prep the lines data
def prep_lines(
    pbp: pd.DataFrame,
    position: Literal["f", "d"] = "f",
    level: Literal["period", "game", "session", "season"] = "game",
    score: bool = False,
    teammates: bool = False,
    opposition: bool = False,
    disable_progress_bar: bool = False,
):
    """Prepares a line stats dataframe using EvolvingHockey data.

    Aggregates to desired level. Capable of returning cuts that account for strength state,
    period, score state, teammates, and opposition.

    Returns a Pandas DataFrame.

    Parameters:
        pbp (pd.DataFrame):
            Dataframe from the prep_pbp function with the default columns argument
        position (str):
            Position to aggregate, forwards or defense, e.g., 'f'
        level (str):
            Level to aggregate stats, e.g., 'game'
        score (bool):
            Whether to aggregate to score state level
        teammates (bool):
            Whether to account for teammates when aggregating
        opposition (bool):
            Whether to account for opposition when aggregating
        disable_progress_bar (bool):
            Whether to disable progress bar

    Returns:
        season (int):
            8-digit season code, e.g., 20232024
        session (str):
            Regular season or playoffs, e.g., R
        game_id (int):
            10-digit game identifier, e.g., 2023020015
        game_date (str):
            Date of game in Eastern time-zone, e.g., 2023-10-12
        team (str):
            3-letter abbreviation of the line's team, e.g., NSH
        opp_team (str):
            3-letter abbreviation of the opposing team, e.g., SEA
        strength_state (str):
            Strength state from the perspective of the event team, e.g., 5v5
        score_state (str):
            Score state from the perspective of the event team, e.g., 0v0
        game_period (int):
            Game period, e.g., 1
        forwards (str):
            Names of the event team's forwards that are on the ice during the event,
            e.g., FILIP.FORSBERG, JUUSO.PARSSINEN, RYAN.O'REILLY
        forwards_id (str):
            EH IDs of the event team's forwards that are on the ice during the event,
            e.g., FILIP.FORSBERG, JUUSO.PARSSINEN, RYAN.O'REILLY
        defense (str):
            Names of the event team's defensemen that are on the ice during the event,
            e.g., ALEX.CARRIER, RYAN.MCDONAGH
        defense_id (str):
            EH IDs of the event team's defensemen that are on the ice during the event,
            e.g., ALEX.CARRIER, RYAN.MCDONAGH
        own_goalie (str):
            Name of the goalie for the event team, e.g., JUUSE.SAROS
        own_goalie_id (str):
            Identifier for the event team goalie that can be used to match with Evolving Hockey data, e.g., JUUSE.SAROS
        opp_forwards (str):
            Names of the opponent's forwards that are on the ice during the event,
            e.g., JARED.MCCANN, JORDAN.EBERLE, MATTY.BENIERS
        opp_forwards_id (str):
            EH IDs of the event team's forwards that are on the ice during the event,
            e.g., JARED.MCCANN, JORDAN.EBERLE, MATTY.BENIERS
        opp_defense (str):
            Names of the opposing team's defensemen that are on the ice during the event,
            e.g., JAMIE.OLEKSIAK, WILLIAM.BORGEN
        opp_defense_id (str):
            EH IDs of the opposing team's defensemen that are on the ice during the event,
            e.g., JAMIE.OLEKSIAK, WILLIAM.BORGEN
        opp_goalie (str):
            Name of the opposing goalie for the event team, e.g., PHILIPP.GRUBAUER
        opp_goalie_id (str):
            Identifier for the opposing goalie that can be used to match with Evolving Hockey data,
            e.g., PHILIPP.GRUBAUER
        toi (float):
            Time on-ice in minutes, e.g., 1.616667
        gf (float):
            Sum of goals scored while line is on-ice, e.g., 0
        gf_adj (float):
            Sum of venue- and score-adjusted goals scored while line is on-ice, e.g., 0
        hdgf (float):
            Sum of high-danger goals scored while line is on-ice, e.g., 0
        ga (float):
            Sum of goals allowed while line is on-ice, e.g., 0
        ga_adj (float):
            Sum of venue- and score-adjusted goals allowed while line is on-ice, e.g., 0
        hdga (float):
            Sum of high-danger goals allowed while line is on-ice, e.g., 0
        xgf (float):
            Sum of expected goals generated while line is on-ice, e.g., 0.017266
        xgf_adj (float):
            Sum of venue- and score-adjusted expected goals generated while line is on-ice, e.g., 0.016472
        xga (float):
            Sum of expected goals allowed while line is on-ice, e.g., 0.123475
        xga_adj (float):
            Sum of venue- and score-adjusted expected goals allowed while line is on-ice, e.g., 0.129772
        sf (float):
            Sum of shots taken while line is on-ice, e.g., 1
        sf_adj (float):
            Sum of venue- and score-adjusted shots taken while line is on-ice, e.g., .972
        hdsf (float):
            Sum of high-danger shots taken while line is on-ice, e.g., 0
        sa (float):
            Sum of shots allowed while line is on-ice, e.g., 0
        sa_adj (float):
            Sum of venue- and score-adjusted shots allowed while line is on-ice, e.g., 0
        hdsa (float):
            Sum of high-danger shots allowed while line is on-ice, e.g., 0
        ff (float):
            Sum of fenwick events generated while line is on-ice, e.g., 1
        ff_adj (float):
            Sum of venue- and score-adjusted fenwick events generated while line is on-ice, e.g., 0.968
        hdff (float):
            Sum of high-danger fenwick events generated while line is on-ice, e.g., 0
        fa (float):
            Sum of fenwick events allowed while line is on-ice, e.g., 1
        fa_adj (float):
            Sum of venue- and score-adjusted fenwick events allowed while line is on-ice, e.g., 1.034
        hdfa (float):
            Sum of high-danger fenwick events allowed while line is on-ice, e.g., 1
        cf (float):
            Sum of corsi events generated while line is on-ice, e.g., 1
        cf_adj (float):
            Sum of venue- and score-adjusted corsi events generated while line is on-ice, e.g., 0.970
        ca (float):
            Sum of corsi events allowed while line is on-ice, e.g., 2
        ca_adj (float):
            Sum of venue- and score-adjusted corsi events allowed while line is on-ice, e.g., 2.064
        bsf (float):
            Sum of shots taken that were ultimately blocked while line is on-ice, e.g., 0
        bsa (float):
            Sum of shots allowed that were ultimately blocked while line is on-ice, e.g., 1
        msf (float):
            Sum of shots taken that missed net while line is on-ice, e.g., 0
        hdmsf (float):
            Sum of high-danger shots taken that missed net while line is on-ice, e.g., 0
        msa (float):
            Sum of shots allowed that missed net while line is on-ice, e.g., 1
        hdmsa (float):
            Sum of high-danger shots allowed that missed net while line is on-ice, e.g., 1
        hf (float):
            Sum of hits dished out while line is on-ice, e.g., 0
        ht (float):
            Sum of hits taken while line is on-ice, e.g., 0
        ozf (float):
            Sum of offensive zone faceoffs that occur while line is on-ice, e.g., 0
        nzf (float):
            Sum of neutral zone faceoffs that occur while line is on-ice, e.g., 0
        dzf (float):
            Sum of defensive zone faceoffs that occur while line is on-ice, e.g., 1
        fow (float):
            Sum of faceoffs won while line is on-ice, e.g., 1
        fol (float):
            Sum of faceoffs lost while line is on-ice, e.g., 0
        ozfw (float):
            Sum of offensive zone faceoffs won while line is on-ice, e.g., 0
        ozfl (float):
            Sum of offensive zone faceoffs lost while line is on-ice, e.g., 1
        nzfw (float):
            Sum of neutral zone faceoffs won while line is on-ice, e.g., 0
        nzfl (float):
            Sum of neutral zone faceoffs lost while line is on-ice, e.g., 0
        dzfw (float):
            Sum of defensive zone faceoffs won while line is on-ice, e.g., 1
        dzfl (float):
            Sum of defensive zone faceoffs lost while line is on-ice, e.g., 0
        pent0 (float):
            Sum of individual 0-minute penalties taken while line is on-ice, e.g., 0
        pent2 (float):
            Sum of individual 2-minute penalties taken while line is on-ice, e.g., 0
        pent4 (float):
            Sum of individual 4-minute penalties taken while line is on-ice, e.g., 0
        pent5 (float):
            Sum of individual 5-minute penalties taken while line is on-ice, e.g., 0
        pent10 (float):
            Sum of individual 10-minute penalties taken while line is on-ice, e.g., 0
        pend0 (float):
            Sum of individual 0-minute penalties drawn while line is on-ice, e.g., 0
        pend2 (float):
            Sum of individual 2-minute penalties drawn while line is on-ice, e.g., 0
        pend4 (float):
            Sum of individual 4-minute penalties drawn while line is on-ice, e.g., 0
        pend5 (float):
            Sum of individual 5-minute penalties drawn while line is on-ice, e.g., 0
        pend10 (float):
            Sum of individual 10-minute penalties drawn while line is on-ice, e.g., 0

    Examples:
        Basic play-by-play DataFrame
        >>> shifts_raw = pd.read_csv("./raw_shifts.csv")
        >>> pbp_raw = pd.read_csv("./raw_pbp.csv")
        >>> pbp = prep_pbp(pbp_raw, shifts_raw)

        Basic game-level stats for forwards, with no teammates or opposition
        >>> lines = prep_lines(pbp, position="f")

        Period-level stats for defense, grouped by teammates
        >>> lines = prep_lines(pbp, position="d", level="period", teammates=True)

        Session-level (e.g., regular seasion) stats, grouped by teammates and opposition
        >>> lines = prep_lines(pbp, position="f", level="session", teammates=True, opposition=True)

    """
    with ChickenProgress(disable=disable_progress_bar) as progress:
        pbar_message = "Prepping lines data..."

        lines_task = progress.add_task(pbar_message, total=1)

        # Creating the "for" dataframe

        # Accounting for desired level of aggregation

        if level == "session" or level == "season":
            group_base = ["season", "session", "event_team", "strength_state"]

        if level == "game":
            group_base = ["season", "game_id", "game_date", "session", "event_team", "opp_team", "strength_state"]

        if level == "period":
            group_base = [
                "season",
                "game_id",
                "game_date",
                "session",
                "event_team",
                "opp_team",
                "game_period",
                "strength_state",
            ]

        # Accounting for score state

        if score is True:
            group_base = group_base + ["score_state"]

        # Accounting for desired position

        group_list = group_base + [f"event_on_{position}", f"event_on_{position}_id"]

        # Accounting for teammates

        if teammates is True:
            if position == "f":
                group_list = group_list + ["event_on_d", "event_on_d_id", "event_on_g", "event_on_g_id"]

            if position == "d":
                group_list = group_list + ["event_on_f", "event_on_f_id", "event_on_g", "event_on_g_id"]

        # Accounting for opposition

        if opposition is True:
            group_list = group_list + ["opp_on_f", "opp_on_f_id", "opp_on_d", "opp_on_d_id", "opp_on_g", "opp_on_g_id"]

            if "opp_team" not in group_list:
                group_list.append("opp_team")

        # Creating dictionary of statistics for the groupby function

        stats = [
            "pred_goal",
            "pred_goal_adj",
            "corsi",
            "corsi_adj",
            "fenwick",
            "fenwick_adj",
            "goal",
            "goal_adj",
            "miss",
            "block",
            "shot",
            "shot_adj",
            "hd_goal",
            "hd_shot",
            "hd_fenwick",
            "hd_miss",
            "event_length",
            "fac",
            "ozf",
            "nzf",
            "dzf",
            "hit",
            "give",
            "take",
            "pen0",
            "pen2",
            "pen4",
            "pen5",
            "pen10",
        ]

        agg_stats = {x: "sum" for x in stats if x in pbp.columns}

        # Aggregating the "for" dataframe

        lines_f = pbp.groupby(group_list, as_index=False, dropna=False).agg(agg_stats)

        # Creating the dictionary to change column names

        columns = [
            "xgf",
            "xgf_adj",
            "cf",
            "cf_adj",
            "ff",
            "ff_adj",
            "gf",
            "gf_adj",
            "msf",
            "bsf",
            "sf",
            "sf_adj",
            "hdgf",
            "hdsf",
            "hdff",
            "hdmsf",
            "toi",
            "fow",
            "ozfw",
            "nzfw",
            "dzfw",
            "hf",
            "give",
            "take",
            "pent0",
            "pent2",
            "pent4",
            "pent5",
            "pent10",
        ]

        columns = dict(zip(stats, columns, strict=False))

        # Accounting for positions

        columns.update(
            {
                "event_on_f": "forwards",
                "event_on_f_id": "forwards_id",
                "event_team": "team",
                "event_on_d": "defense",
                "event_on_d_id": "defense_id",
                "event_on_g": "own_goalie",
                "event_on_g_id": "own_goalie_id",
                "opp_on_f": "opp_forwards",
                "opp_on_f_id": "opp_forwards_id",
                "opp_on_d": "opp_defense",
                "opp_on_d_id": "opp_defense_id",
                "opp_on_g": "opp_goalie",
                "opp_on_g_id": "opp_goalie_id",
            }
        )

        # columns = {k: v for k, v in columns.items() if k in lines_f.columns}

        lines_f = lines_f.rename(columns=columns)

        cols = [
            "forwards",
            "forwards_id",
            "defense",
            "defense_id",
            "own_goalie",
            "own_goalie_id",
            "opp_forwards",
            "opp_forwards_id",
            "opp_defense",
            "opp_defense_id",
            "opp_goalie",
            "opp_goalie_id",
        ]

        cols = [x for x in cols if x in lines_f]

        for col in cols:
            lines_f[col] = lines_f[col].fillna("EMPTY")

        # Creating the against dataframe

        # Accounting for desired level of aggregation

        if level == "session" or level == "season":
            group_base = ["season", "session", "opp_team", "opp_strength_state"]

        if level == "game":
            group_base = ["season", "game_id", "game_date", "session", "event_team", "opp_team", "opp_strength_state"]

        if level == "period":
            group_base = [
                "season",
                "game_id",
                "game_date",
                "session",
                "event_team",
                "opp_team",
                "game_period",
                "opp_strength_state",
            ]

        # Accounting for score state

        if score is True:
            group_base = group_base + ["opp_score_state"]

        # Accounting for desired position

        group_list = group_base + [f"opp_on_{position}", f"opp_on_{position}_id"]

        # Accounting for teammates

        if teammates is True:
            if position == "f":
                group_list = group_list + ["opp_on_d", "opp_on_d_id", "opp_on_g", "opp_on_g_id"]

            if position == "d":
                group_list = group_list + ["opp_on_f", "opp_on_f_id", "opp_on_g", "opp_on_g_id"]

        # Accounting for opposition

        if opposition is True:
            group_list = group_list + [
                "event_on_f",
                "event_on_f_id",
                "event_on_d",
                "event_on_d_id",
                "event_on_g",
                "event_on_g_id",
            ]

            if "event_team" not in group_list:
                group_list.append("event_team")

        # Creating dictionary of statistics for the groupby function

        stats = [
            "pred_goal",
            "pred_goal_adj",
            "corsi",
            "corsi_adj",
            "fenwick",
            "fenwick_adj",
            "goal",
            "goal_adj",
            "miss",
            "block",
            "shot",
            "shot_adj",
            "hd_goal",
            "hd_shot",
            "hd_fenwick",
            "hd_miss",
            "event_length",
            "fac",
            "ozf",
            "nzf",
            "dzf",
            "hit",
            "pen0",
            "pen2",
            "pen4",
            "pen5",
            "pen10",
        ]

        agg_stats = {x: "sum" for x in stats if x in pbp.columns}

        # Aggregating "against" dataframe

        lines_a = pbp.groupby(group_list, as_index=False, dropna=False).agg(agg_stats)

        # Creating the dictionary to change column names

        columns = [
            "xga",
            "xga_adj",
            "ca",
            "ca_adj",
            "fa",
            "fa_adj",
            "ga",
            "ga_adj",
            "msa",
            "bsa",
            "sa",
            "sa_adj",
            "hdga",
            "hdsa",
            "hdfa",
            "hdmsa",
            "toi",
            "fol",
            "ozfl",
            "nzfl",
            "dzfl",
            "ht",
            "pend0",
            "pend2",
            "pend4",
            "pend5",
            "pend10",
        ]

        columns = dict(zip(stats, columns, strict=False))

        # Accounting for positions

        columns.update(
            {
                "opp_team": "team",
                "event_team": "opp_team",
                "opp_on_f": "forwards",
                "opp_on_f_id": "forwards_id",
                "opp_strength_state": "strength_state",
                "opp_on_d": "defense",
                "opp_on_d_id": "defense_id",
                "event_on_f": "opp_forwards",
                "event_on_f_id": "opp_forwards_id",
                "event_on_d": "opp_defense",
                "event_on_d_id": "opp_defense_id",
                "opp_score_state": "score_state",
                "event_on_g": "opp_goalie",
                "event_on_g_id": "opp_goalie_id",
                "opp_on_g": "own_goalie",
                "opp_on_g_id": "own_goalie_id",
            }
        )

        # columns = {k: v for k, v in columns.items() if k in lines_a.columns}

        lines_a = lines_a.rename(columns=columns)

        cols = [
            "forwards",
            "forwards_id",
            "defense",
            "defense_id",
            "own_goalie",
            "own_goalie_id",
            "opp_forwards",
            "opp_forwards_id",
            "opp_defense",
            "opp_defense_id",
            "opp_goalie",
            "opp_goalie_id",
        ]

        cols = [x for x in cols if x in lines_a]

        for col in cols:
            lines_a[col] = lines_a[col].fillna("EMPTY")

        # Merging the "for" and "against" dataframes

        if level == "session" or level == "season":
            if position == "f":
                merge_list = ["season", "session", "team", "strength_state", "forwards", "forwards_id"]

            if position == "d":
                merge_list = ["season", "session", "team", "strength_state", "defense", "defense_id"]

        if level == "game":
            if position == "f":
                merge_list = [
                    "season",
                    "game_id",
                    "game_date",
                    "session",
                    "team",
                    "opp_team",
                    "strength_state",
                    "forwards",
                    "forwards_id",
                ]

            if position == "d":
                merge_list = [
                    "season",
                    "game_id",
                    "game_date",
                    "session",
                    "team",
                    "opp_team",
                    "strength_state",
                    "defense",
                    "defense_id",
                ]

        if level == "period":
            if position == "f":
                merge_list = [
                    "season",
                    "game_id",
                    "game_date",
                    "session",
                    "team",
                    "opp_team",
                    "strength_state",
                    "forwards",
                    "forwards_id",
                    "game_period",
                ]

            if position == "d":
                merge_list = [
                    "season",
                    "game_id",
                    "game_date",
                    "session",
                    "team",
                    "opp_team",
                    "strength_state",
                    "defense",
                    "defense_id",
                    "game_period",
                ]

        if score is True:
            merge_list.append("score_state")

        if teammates is True:
            if position == "f":
                merge_list = merge_list + ["defense", "defense_id", "own_goalie", "own_goalie_id"]

            if position == "d":
                merge_list = merge_list + ["forwards", "forwards_id", "own_goalie", "own_goalie_id"]

        if opposition is True:
            merge_list = merge_list + [
                "opp_forwards",
                "opp_forwards_id",
                "opp_defense",
                "opp_defense_id",
                "opp_goalie",
                "opp_goalie_id",
            ]

            if "opp_team" not in merge_list:
                merge_list.insert(3, "opp_team")

        lines = lines_f.merge(lines_a, how="outer", on=merge_list, suffixes=("_x", "")).fillna(0)

        lines.toi = (lines.toi_x + lines.toi) / 60

        lines = lines.drop(columns="toi_x")

        lines["ozf"] = lines.ozfw + lines.ozfl

        lines["nzf"] = lines.nzfw + lines.nzfl

        lines["dzf"] = lines.dzfw + lines.dzfl

        cols = [x for x in LineSchema.dtypes if x in lines.columns]

        lines = lines[cols]

        lines = lines.loc[lines.toi > 0].reset_index(drop=True).copy()

        lines = LineSchema.validate(lines)

        pbar_message = "Finished prepping lines data"

        progress.update(lines_task, description=pbar_message, advance=1, refresh=True)

    return lines


# Function to prep the team stats
def prep_team(
    pbp: pd.DataFrame,
    level: Literal["period", "game", "session", "season"] = "game",
    strengths: bool = True,
    score: bool = False,
    disable_progress_bar: bool = False,
) -> pd.DataFrame:
    """Prepares a team stats dataframe using Evolving Hockey data.

    Aggregates to desired level. Capable of returning cuts that account for strength state,
    period, and score state. Returns a Pandas DataFrame.

    Parameters:
        pbp (pd.DataFrame):
            Dataframe from the prep_pbp function with the default columns argument
        level (str):
            Level to aggregate stats, e.g., 'game'
        strengths (bool):
            Whether to aggregate to strength state level, e.g., True
        score (bool):
            Whether to aggregate to score state level
        disable_progress_bar (bool):
            Whether to disable progress bar

    Returns:
        season (int):
            8-digit season code, e.g., 20232024
        session (str):
            Regular season or playoffs, e.g., R
        game_id (int):
            10-digit game identifier, e.g., 2023020044
        game_date (str):
            Date of game in Eastern time-zone, e.g., 2023-10-17
        team (str):
            3-letter abbreviation of the team, e.g., NSH
        opp_team (str):
            3-letter abbreviation of the opposing team, e.g., EDM
        strength_state (str):
            Strength state from the perspective of the event team, e.g., 5v5
        score_state (str):
            Score state from the perspective of the event team, e.g., 1v6
        game_period (int):
            Game period, e.g., 3
        toi (float):
            Time on-ice in minutes, e.g., 18
        gf (float):
            Sum of goals scored, e.g., 0
        gf_adj (float):
            Sum of venue- and score-adjusted goals scored, e.g., 0
        hdgf (float):
            Sum of high-danger goals scored, e.g., 0
        ga (float):
            Sum of goals allowed, e.g., 0
        ga_adj (float):
            Sum of venue- and score-adjusted goals allowed, e.g., 0
        hdga (float):
            Sum of high-danger goals allowed, e.g., 0
        xgf (float):
            Sum of expected goals generated, e.g., 0.957070
        xgf_adj (float):
            Sum of venue- and score-adjusted expected goals generated, e.g., 0.883376
        xga (float):
            Sum of expected goals allowed, e.g., 0.535971
        xga_adj (float):
            Sum of venue- and score-adjusted expected goals allowed, e.g., 0.584744
        sf (float):
            Sum of shots taken, e.g., 10
        sf_adj (float):
            Sum of venue- and score-adjusted shots taken, e.g., 8.620
        hdsf (float):
            Sum of high-danger shots taken, e.g., 2
        sa (float):
            Sum of shots allowed, e.g., 4
        sa_adj (float):
            Sum of venue- and score-adjusted shots allowed, e.g., 4.764
        hdsa (float):
            Sum of high-danger shots allowed, e.g., 0
        ff (float):
            Sum of fenwick events generated, e.g., 14
        ff_adj (float):
            Sum of venue- and score-adjusted fenwick events generated, e.g., 12.026
        hdff (float):
            Sum of high-danger fenwick events generated, e.g., 2
        fa (float):
            Sum of fenwick events allowed, e.g., 8
        fa_adj (float):
            Sum of venue- and score-adjusted fenwick events allowed, e.g., 9.576
        hdfa (float):
            Sum of high-danger fenwick events allowed, e.g., 1
        cf (float):
            Sum of corsi events generated, e.g., 16
        cf_adj (float):
            Sum of venue- and score-adjusted corsi events generated, e.g., 13.488
        ca (float):
            Sum of corsi events allowed, e.g., 12.0
        ca_adj (float):
            Sum of venue- and score-adjusted corsi events allowed, e.g., 14.760
        bsf (float):
            Sum of shots taken that were ultimately blocked, e.g., 4
        bsa (float):
            Sum of shots allowed that were ultimately blocked, e.g., 2
        msf (float):
            Sum of shots taken that missed net, e.g., 4
        hdmsf (float):
            Sum of high-danger shots taken that missed net, e.g., 0
        msa (float):
            Sum of shots allowed that missed net, e.g., 4
        hdmsa (float):
            Sum of high-danger shots allowed that missed net, e.g., 1
        ozf (float):
            Sum of offensive zone faceoffs that occur, e.g., 6
        nzf (float):
            Sum of neutral zone faceoffs that occur, e.g., 4
        dzf (float):
            Sum of defensive zone faceoffs that occur, e.g., 6
        fow (float):
            Sum of faceoffs won, e.g., 8
        fol (float):
            Sum of faceoffs lost, e.g., 11
        ozfw (float):
            Sum of offensive zone faceoffs won, e.g., 3
        ozfl (float):
            Sum of offensive zone faceoffs lost, e.g., 1
        nzfw (float):
            Sum of neutral zone faceoffs won, e.g., 2
        nzfl (float):
            Sum of neutral zone faceoffs lost, e.g., 3
        dzfw (float):
            Sum of defensive zone faceoffs won, e.g., 3
        dzfl (float):
            Sum of defensive zone faceoffs lost, e.g., 7
        hf (float):
            Sum of hits dished out, e.g., 7
        ht (float):
            Sum of hits taken, e.g., 5
        give (float):
            Sum of giveaways, e.g., 5
        take (float):
            Sum of takeaways, e.g., 1
        pent0 (float):
            Sum of individual 0-minute penalties taken, e.g., 0
        pent2 (float):
            Sum of individual 2-minute penalties taken, e.g., 0
        pent4 (float):
            Sum of individual 4-minute penalties taken, e.g., 0
        pent5 (float):
            Sum of individual 5-minute penalties taken, e.g., 0
        pent10 (float):
            Sum of individual 10-minute penalties taken, e.g., 0
        pend0 (float):
            Sum of individual 0-minute penalties drawn, e.g., 0
        pend2 (float):
            Sum of individual 2-minute penalties drawn, e.g., 0
        pend4 (float):
            Sum of individual 4-minute penalties drawn, e.g., 0
        pend5 (float):
            Sum of individual 5-minute penalties drawn, e.g., 0
        pend10 (float):
            Sum of individual 10-minute penalties drawn, e.g., 0

    Examples:
        Basic play-by-play DataFrame
        >>> shifts_raw = pd.read_csv("./raw_shifts.csv")
        >>> pbp_raw = pd.read_csv("./raw_pbp.csv")
        >>> pbp = prep_pbp(pbp_raw, shifts_raw)

        Basic game-level stats for teams
        >>> team = prep_team(pbp)

        Period-level team stats, grouped by score state
        >>> team = prep_team(pbp, level="period", score=True)
    """
    with ChickenProgress(disable=disable_progress_bar) as progress:
        pbar_message = "Prepping team data..."

        team_task = progress.add_task(pbar_message, total=1)

        # Getting the "for" stats

        group_list = ["season", "session", "event_team"]

        if strengths is True:
            group_list.append("strength_state")

        if level == "game" or level == "period":
            group_list.insert(3, "opp_team")

            group_list[2:2] = ["game_id", "game_date"]

        if level == "period":
            group_list.append("game_period")

        if score is True:
            group_list.append("score_state")

        agg_stats = [
            "pred_goal",
            "pred_goal_adj",
            "shot",
            "shot_adj",
            "miss",
            "block",
            "corsi",
            "corsi_adj",
            "fenwick",
            "fenwick_adj",
            "goal",
            "goal_adj",
            "give",
            "take",
            "hd_goal",
            "hd_shot",
            "hd_fenwick",
            "hd_miss",
            "hit",
            "pen0",
            "pen2",
            "pen4",
            "pen5",
            "pen10",
            "fac",
            "ozf",
            "nzf",
            "dzf",
            "event_length",
        ]

        agg_dict = {x: "sum" for x in agg_stats if x in pbp.columns}

        new_cols = [
            "xgf",
            "xgf_adj",
            "sf",
            "sf_adj",
            "msf",
            "bsa",
            "cf",
            "cf_adj",
            "ff",
            "ff_adj",
            "gf",
            "gf_adj",
            "give",
            "take",
            "hdgf",
            "hdsf",
            "hdff",
            "hdmsf",
            "hf",
            "pent0",
            "pent2",
            "pent4",
            "pent5",
            "pent10",
            "fow",
            "ozfw",
            "nzfw",
            "dzfw",
            "toi",
        ]

        new_cols = dict(zip(agg_stats, new_cols, strict=False))

        new_cols.update({"event_team": "team"})

        stats_for = pbp.groupby(group_list, as_index=False).agg(agg_dict).rename(columns=new_cols)

        # Getting the "against" stats

        group_list = ["season", "session", "opp_team"]

        if strengths is True:
            group_list.append("opp_strength_state")

        if level == "game" or level == "period":
            group_list.insert(3, "event_team")

            group_list[2:2] = ["game_id", "game_date"]

        if level == "period":
            group_list.append("game_period")

        if score is True:
            group_list.append("opp_score_state")

        agg_stats = [
            "pred_goal",
            "pred_goal_adj",
            "shot",
            "shot_adj",
            "miss",
            "block",
            "corsi",
            "corsi_adj",
            "fenwick",
            "fenwick_adj",
            "goal",
            "goal_adj",
            "hd_goal",
            "hd_shot",
            "hd_fenwick",
            "hd_miss",
            "hit",
            "pen0",
            "pen2",
            "pen4",
            "pen5",
            "pen10",
            "fac",
            "ozf",
            "nzf",
            "dzf",
            "event_length",
        ]

        agg_dict = {x: "sum" for x in agg_stats if x in pbp.columns}

        new_cols = [
            "xga",
            "xga_adj",
            "sa",
            "sa_adj",
            "msa",
            "bsf",
            "ca",
            "ca_adj",
            "fa",
            "fa_adj",
            "ga",
            "ga_adj",
            "hdga",
            "hdsa",
            "hdfa",
            "hdmsa",
            "ht",
            "pend0",
            "pend2",
            "pend4",
            "pend5",
            "pend10",
            "fol",
            "ozfl",
            "nzfl",
            "dzfl",
            "toi",
        ]

        new_cols = dict(zip(agg_stats, new_cols, strict=False))

        new_cols.update(
            {
                "opp_team": "team",
                "opp_score_state": "score_state",
                "opp_strength_state": "strength_state",
                "event_team": "opp_team",
            }
        )

        stats_against = pbp.groupby(group_list, as_index=False).agg(agg_dict).rename(columns=new_cols)

        merge_list = [
            "season",
            "session",
            "game_id",
            "game_date",
            "team",
            "opp_team",
            "strength_state",
            "score_state",
            "game_period",
        ]

        merge_list = [x for x in merge_list if x in stats_for.columns and x in stats_against.columns]

        team_stats = stats_for.merge(stats_against, on=merge_list, how="outer")

        team_stats["toi"] = (team_stats.toi_x + team_stats.toi_y) / 60

        team_stats = team_stats.drop(["toi_x", "toi_y"], axis=1)

        fos = ["ozf", "nzf", "dzf"]

        for fo in fos:
            team_stats[fo] = team_stats[f"{fo}w"] + team_stats[f"{fo}w"]

        team_stats = team_stats.dropna(subset="toi").reset_index(drop=True)

        cols = [x for x in TeamStatSchema.dtypes if x in team_stats.columns]

        team_stats = TeamStatSchema.validate(team_stats[cols])

        pbar_message = "Finished prepping team data"

        progress.update(team_task, description=pbar_message, advance=1, refresh=True)

    return team_stats


# Function to prep the GAR dataframe
def prep_gar(skater_data: pd.DataFrame, goalie_data: pd.DataFrame) -> pd.DataFrame:
    """Prepares a dataframe of GAR stats using Evolving Hockey data.

    Experimental and not actively maintained

    Parameters:
        skater_data (pd.DataFrame):
            Pandas Dataframe loaded from a CSV file from Evolving Hockey website
        goalie_data (pd.DataFrame):
            Pandas Dataframe loaded from a CSV file from Evolving Hockey website

    """
    gar = pd.concat([skater_data, goalie_data], ignore_index=True)

    new_cols = {x: x.replace(" ", "_").lower() for x in gar.columns}

    gar = gar.rename(columns=new_cols)

    season_split = gar.season.str.split("-", expand=True)

    gar.season = "20" + season_split[0] + "20" + season_split[1]

    gar.birthday = pd.to_datetime(gar.birthday)

    gar.player = gar.player.str.upper()

    gar.eh_id = gar.eh_id.str.replace("..", ".", regex=False)

    replace_teams = {"S.J": "SJS", "N.J": "NJD", "T.B": "TBL", "L.A": "LAK"}

    gar.team = gar.team.map(replace_teams).fillna(gar.team)

    gar = gar.rename(columns={"eh_id": "player_id"})

    return gar


# Function to prep the xGAR dataframe
def prep_xgar(data: pd.DataFrame) -> pd.DataFrame:
    """Prepares a dataframe of xGAR stats using Evolving Hockey data.

    Experimental and not actively maintained

    Parameters:
        data (pd.DataFrame):
            Pandas Dataframe loaded from a CSV file from Evolving Hockey website

    """
    xgar = data.copy()

    new_cols = {x: x.replace(" ", "_").lower() for x in xgar.columns}

    xgar = xgar.rename(columns=new_cols)

    season_split = xgar.season.str.split("-", expand=True)

    xgar.season = "20" + season_split[0] + "20" + season_split[1]

    xgar.birthday = pd.to_datetime(xgar.birthday)

    xgar.player = xgar.player.str.upper()

    xgar.eh_id = xgar.eh_id.str.replace("..", ".", regex=False)

    replace_teams = {"S.J": "SJS", "N.J": "NJD", "T.B": "TBL", "L.A": "LAK"}

    xgar.team = xgar.team.map(replace_teams).fillna(xgar.team)

    xgar = xgar.rename(columns={"eh_id": "player_id"})

    return xgar



import importlib.resources

import matplotlib.pyplot as plt
import requests
import urllib3
from matplotlib import _rc_params_in_file
from requests.adapters import HTTPAdapter
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    ProgressColumn,
    SpinnerColumn,
    Task,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)
from rich.text import Text


class ChickenHTTPAdapter(HTTPAdapter):
    """Modified HTTPAdapter for managing requests timeouts."""

    def __init__(self, *args, **kwargs):
        """Initializes HTTPAdapter for managing requests timeouts."""
        self.timeout = 3

        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]

            del kwargs["timeout"]

        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        """Modifies the HTTPAdapter's send method to manage requests timeouts."""
        timeout = kwargs.get("timeout")

        if timeout is None:
            kwargs["timeout"] = self.timeout

        return super().send(request, **kwargs)


class ChickenSession(requests.Session):
    """Modified Requests session for use with chickenstats library."""

    def __init__(self):
        """Initializes Requests Session object."""
        super().__init__()

        retry = urllib3.Retry(
            total=10,
            backoff_factor=2,
            respect_retry_after_header=False,
            status_forcelist=[54, 60, 401, 403, 404, 408, 429, 500, 502, 503, 504],
        )

        connect_timeout = 3
        read_timeout = 10

        adapter = ChickenHTTPAdapter(max_retries=retry, timeout=(connect_timeout, read_timeout))

        self.mount("http://", adapter)
        self.mount("https://", adapter)


class ScrapeSpeedColumn(ProgressColumn):
    """Renders human-readable transfer speed."""

    def render(self, task: "Task") -> Text:
        """Show data transfer speed."""
        speed = task.finished_speed or task.speed
        if speed is None:
            return Text("?", style="progress.data.speed")
        else:
            speed = round(speed, 2)

            if speed < 1:
                speed = round(1 / speed, 2)
                pbar_text = f"{speed:.2f} s/it"

            else:
                pbar_text = f"{speed:.2f} it/s"

        return Text(pbar_text, style="progress.data.speed")


class ChickenProgress(Progress):
    """Progress bar to be used across modules."""

    def __init__(self, disable: bool = False):
        """Progress bar to be used across modules."""
        super().__init__(
            TextColumn("[progress.description]{task.description}"),
            SpinnerColumn(),
            BarColumn(),
            TaskProgressColumn(),
            TextColumn("•"),
            TimeElapsedColumn(),
            TextColumn("•"),
            TimeRemainingColumn(),
            TextColumn("•"),
            MofNCompleteColumn(),
            TextColumn("•"),
            ScrapeSpeedColumn(),
            disable=disable,
        )


class ChickenProgressIndeterminate(Progress):
    """Indeterminate progress bar to be used across modules."""

    def __init__(self, disable: bool = False):
        """Progress bar to be used across modules."""
        super().__init__(
            TextColumn("[progress.description]{task.description}"),
            SpinnerColumn(),
            BarColumn(),
            TextColumn("•"),
            TimeElapsedColumn(),
            disable=disable,
        )


def add_cs_mplstyles():
    """Add chickenstats matplotlib style to style library for later usage."""
    styles = dict()

    with importlib.resources.as_file(
        importlib.resources.files("chickenstats.utilities.styles").joinpath("chickenstats.mplstyle")
    ) as file:
        styles["chickenstats"] = _rc_params_in_file(file)

    with importlib.resources.as_file(
        importlib.resources.files("chickenstats.utilities.styles").joinpath("chickenstats_dark.mplstyle")
    ) as file:
        styles["chickenstats_dark"] = _rc_params_in_file(file)

    # Update dictionary of styles
    plt.style.core.update_nested_dict(plt.style.library, styles)
    plt.style.core.available[:] = sorted(plt.style.library.keys())



from requests import Session
from requests.adapters import HTTPAdapter
from rich.progress import Progress

from chickenstats.utilities.utilities import ChickenHTTPAdapter, ChickenProgress, ChickenSession


def test_ChickenProgress():
    progress = ChickenProgress()

    assert isinstance(progress, Progress) is True


def test_ChickenSession():
    session = ChickenSession()

    assert isinstance(session, Session) is True


def test_ChickenHTTPAdapter():
    adapter = ChickenHTTPAdapter()

    assert isinstance(adapter, HTTPAdapter) is True



from chickenstats.utilities.utilities import (ChickenProgress,
                                              ChickenProgressIndeterminate,
                                              ChickenSession,
                                              add_cs_mplstyles)

add_cs_mplstyles()



# from pathlib import Path

import numpy as np
import pandas as pd

# import mlflow
# from sklearn.model_selection import train_test_split
#
# import matplotlib.pyplot as plt
# import shap
# from yellowbrick.classifier import (
#     ClassificationReport,
#     ClassPredictionError,
#     ROCAUC,
#     PrecisionRecallCurve,
#     ConfusionMatrix,
# )
# from yellowbrick.model_selection import FeatureImportances
from chickenstats.chicken_nhl._validation import XGSchema


def prep_data(data: pd.DataFrame, strengths: str) -> pd.DataFrame:
    """Function for prepping play-by-play data for xG experiments.

    Data are play-by-play data from the chickenstats function.

    Strengths can be: even, powerplay, shorthanded, empty_for, empty_against
    """
    df = data.copy()

    events = [
        "SHOT",
        "FAC",
        "HIT",
        "BLOCK",
        "MISS",
        "GIVE",
        "TAKE",
        # "PENL",
        "GOAL",
    ]

    conds = np.logical_and.reduce(
        [df.event.isin(events), df.strength_state != "1v0", pd.notnull(df.coords_x), pd.notnull(df.coords_y)]
    )

    df = df.loc[conds]

    conds = np.logical_and.reduce(
        [df.season == df.season.shift(1), df.game_id == df.game_id.shift(1), df.period == df.period.shift(1)]
    )
    df["seconds_since_last"] = np.where(conds, df.game_seconds - df.game_seconds.shift(1), np.nan)
    df["event_type_last"] = np.where(conds, df.event.shift(1), np.nan)
    df["event_team_last"] = np.where(conds, df.event_team.shift(1), np.nan)
    df["event_strength_last"] = np.where(conds, df.strength_state.shift(1), np.nan)
    df["coords_x_last"] = np.where(conds, df.coords_x.shift(1), np.nan)
    df["coords_y_last"] = np.where(conds, df.coords_y.shift(1), np.nan)
    df["zone_last"] = np.where(conds, df.zone.shift(1), np.nan)

    df["same_team_last"] = np.where(np.equal(df.event_team, df.event_team_last), 1, 0)

    df["distance_from_last"] = ((df.coords_x - df.coords_x_last) ** 2 + (df.coords_y - df.coords_y_last) ** 2) ** (
        1 / 2
    )

    last_is_shot = np.equal(df.event_type_last, "SHOT")
    last_is_miss = np.equal(df.event_type_last, "MISS")
    last_is_block = np.equal(df.event_type_last, "BLOCK")
    last_is_give = np.equal(df.event_type_last, "GIVE")
    last_is_take = np.equal(df.event_type_last, "TAKE")
    last_is_hit = np.equal(df.event_type_last, "HIT")
    last_is_fac = np.equal(df.event_type_last, "FAC")

    same_team_as_last = np.equal(df.same_team_last, 1)
    not_same_team_as_last = np.equal(df.same_team_last, 0)

    df["prior_shot_same"] = np.where((last_is_shot & same_team_as_last), 1, 0)
    df["prior_miss_same"] = np.where((last_is_miss & same_team_as_last), 1, 0)
    df["prior_block_same"] = np.where((last_is_block & same_team_as_last), 1, 0)
    df["prior_give_same"] = np.where((last_is_give & same_team_as_last), 1, 0)
    df["prior_take_same"] = np.where((last_is_take & same_team_as_last), 1, 0)
    df["prior_hit_same"] = np.where((last_is_hit & same_team_as_last), 1, 0)

    df["prior_shot_opp"] = np.where((last_is_shot & not_same_team_as_last), 1, 0)
    df["prior_miss_opp"] = np.where((last_is_miss & not_same_team_as_last), 1, 0)
    df["prior_block_opp"] = np.where((last_is_block & not_same_team_as_last), 1, 0)
    df["prior_give_opp"] = np.where((last_is_give & not_same_team_as_last), 1, 0)
    df["prior_take_opp"] = np.where((last_is_take & not_same_team_as_last), 1, 0)
    df["prior_hit_opp"] = np.where((last_is_hit & not_same_team_as_last), 1, 0)

    df["prior_face"] = np.where(last_is_fac, 1, 0)

    shot_types = pd.get_dummies(df.shot_type, dtype=int)

    shot_types = shot_types.rename(
        columns={x: x.lower().replace("-", "_").replace(" ", "_") for x in shot_types.columns}
    )

    df = df.copy().merge(shot_types, left_index=True, right_index=True, how="outer")

    conds = [df.score_diff > 4, df.score_diff < -4]

    values = [4, -4]

    df.score_diff = np.select(conds, values, df.score_diff)

    conds = [df.player_1_position.isin(["F", "L", "R", "C"]), df.player_1_position == "D", df.player_1_position == "G"]

    values = ["F", "D", "G"]

    df["position_group"] = np.select(conds, values, default="F")

    position_dummies = pd.get_dummies(df.position_group, dtype=int)

    new_cols = {x: f"position_{x.lower()}" for x in values}

    position_dummies = position_dummies.rename(columns=new_cols)

    df = df.merge(position_dummies, left_index=True, right_index=True)

    conds = [
        np.logical_and.reduce(
            [
                df.event.isin(["GOAL", "SHOT", "BLOCK", "MISS"]),
                df.event_type_last.isin(["SHOT", "MISS"]),
                df.event_team_last == df.event_team,
                df.game_id == df.game_id.shift(1),
                df.period == df.period.shift(1),
                df.seconds_since_last <= 3,
            ]
        ),
        np.logical_and.reduce(
            [
                df.event.isin(["GOAL", "SHOT", "BLOCK", "MISS"]),
                df.event_type_last == "BLOCK",
                df.event_team_last == df.opp_team,
                df.game_id == df.game_id.shift(1),
                df.period == df.period.shift(1),
                df.seconds_since_last <= 3,
            ]
        ),
    ]

    values = [1, 1]

    df["is_rebound"] = np.select(conds, values, 0)

    conds = np.logical_and.reduce(
        [
            df.event.isin(["GOAL", "SHOT", "BLOCK", "MISS"]),
            df.seconds_since_last <= 4,
            df.zone_last == "NEU",
            df.game_id == df.game_id.shift(1),
            df.period == df.period.shift(1),
            df.event != "FAC",
        ]
    )

    df["rush_attempt"] = np.where(conds, 1, 0)

    cat_cols = ["strength_state", "position_group", "event_type_last"]

    for col in cat_cols:
        dummies = pd.get_dummies(df[col], dtype=int)

        new_cols = {x: f"{col}_{x}" for x in dummies.columns}

        dummies = dummies.rename(columns=new_cols)

        df = df.copy().merge(dummies, left_index=True, right_index=True)

    if strengths.lower() == "even":
        strengths_list = ["5v5", "4v4", "3v3"]

    if strengths.lower() == "powerplay" or strengths.lower() == "pp":
        strengths_list = ["5v4", "4v3", "5v3"]

    if strengths.lower() == "shorthanded" or strengths.lower() == "ss":
        strengths_list = ["4v5", "3v4", "3v5"]

    if strengths.lower() == "empty_for":
        strengths_list = ["Ev5", "Ev4", "Ev3"]

    if strengths.lower() == "empty_against":
        strengths_list = ["5vE", "4vE", "3vE"]

    conds = np.logical_and.reduce([df.event.isin(["GOAL", "SHOT", "MISS"]), df.strength_state.isin(strengths_list)])

    df = df.loc[conds]

    drop_cols = [
        x for x in df.columns if "strength_state_" in x and x not in [f"strength_state_{x}" for x in strengths_list]
    ] + cat_cols

    df = df.drop(drop_cols, axis=1, errors="ignore")

    df = XGSchema.validate(df[[x for x in XGSchema.dtypes if x in df.columns]])

    return df



from pandera.pandas import Column, DataFrameSchema, Index

PBPSchema = DataFrameSchema(
    columns={
        "id": Column(int),
        "season": Column(int),
        "session": Column(str),
        "game_id": Column(int),
        "game_date": Column(str),
        "event_index": Column(int),
        "game_period": Column(int),
        "period_seconds": Column(int),
        "game_seconds": Column(int),
        "clock_time": Column(str),
        "strength_state": Column(str, nullable=True),
        "event_team": Column(str, nullable=True),
        "opp_team": Column(str, nullable=True),
        "event_type": Column(str),
        "event_description": Column(str, nullable=True),
        "event_detail": Column(str, nullable=True),
        "event_zone": Column(str, nullable=True),
        "coords_x": Column(float, nullable=True),
        "coords_y": Column(float, nullable=True),
        "high_danger": Column(int, default=0),
        "danger": Column(int, default=0),
        "event_player_1": Column(str, nullable=True),
        "event_player_1_id": Column(str, nullable=True),
        "event_player_1_pos": Column(str, nullable=True),
        "event_player_2": Column(str, nullable=True),
        "event_player_2_id": Column(str, nullable=True),
        "event_player_2_pos": Column(str, nullable=True),
        "event_player_3": Column(str, nullable=True),
        "event_player_3_id": Column(str, nullable=True),
        "event_player_3_pos": Column(str, nullable=True),
        "score_state": Column(str, nullable=True),
        "event_length": Column(int, default=0),
        "event_distance": Column(float, nullable=True),
        "pbp_distance": Column(float, nullable=True),
        "event_angle": Column(float, nullable=True),
        "home_score": Column(int, required=False),
        "away_score": Column(int, required=False),
        "is_home": Column(float, nullable=True),
        "home_team": Column(str, required=False),
        "away_team": Column(str, required=False),
        "home_skaters": Column(int, required=False),
        "away_skaters": Column(int, required=False),
        "home_zone": Column(str, required=False, nullable=True),
        "home_goalie": Column(str, required=False),
        "away_goalie": Column(str, required=False),
        "home_zonestart": Column(float, required=False, nullable=True),
        "face_index": Column(int, required=False),
        "pen_index": Column(int, required=False),
        "shift_index": Column(int, required=False),
        "game_score_state": Column(str, nullable=True, default=None),
        "game_strength_state": Column(str, nullable=True, default=None),
        "event_on_f": Column(str, nullable=True),
        "event_on_f_id": Column(str, nullable=True),
        "event_on_d": Column(str, nullable=True),
        "event_on_d_id": Column(str, nullable=True),
        "event_on_g": Column(str, nullable=True),
        "event_on_g_id": Column(str, nullable=True),
        "event_on_1": Column(str, required=False, nullable=True),
        "event_on_1_id": Column(str, required=False, nullable=True),
        "event_on_1_pos": Column(str, required=False, nullable=True),
        "event_on_2": Column(str, required=False, nullable=True),
        "event_on_2_id": Column(str, required=False, nullable=True),
        "event_on_2_pos": Column(str, required=False, nullable=True),
        "event_on_3": Column(str, required=False, nullable=True),
        "event_on_3_id": Column(str, required=False, nullable=True),
        "event_on_3_pos": Column(str, required=False, nullable=True),
        "event_on_4": Column(str, required=False, nullable=True),
        "event_on_4_id": Column(str, required=False, nullable=True),
        "event_on_4_pos": Column(str, required=False, nullable=True),
        "event_on_5": Column(str, required=False, nullable=True),
        "event_on_5_id": Column(str, required=False, nullable=True),
        "event_on_5_pos": Column(str, required=False, nullable=True),
        "event_on_6": Column(str, required=False, nullable=True),
        "event_on_6_id": Column(str, required=False, nullable=True),
        "event_on_6_pos": Column(str, required=False, nullable=True),
        "event_on_7": Column(str, required=False, nullable=True),
        "event_on_7_id": Column(str, required=False, nullable=True),
        "event_on_7_pos": Column(str, required=False, nullable=True),
        "opp_strength_state": Column(str, nullable=True, default=None),
        "opp_score_state": Column(str, nullable=True, default=None),
        "opp_on_f": Column(str, nullable=True),
        "opp_on_f_id": Column(str, nullable=True),
        "opp_on_d": Column(str, nullable=True),
        "opp_on_d_id": Column(str, nullable=True),
        "opp_on_g": Column(str, nullable=True),
        "opp_on_g_id": Column(str, nullable=True),
        "opp_on_1": Column(str, required=False, nullable=True),
        "opp_on_1_id": Column(str, required=False, nullable=True),
        "opp_on_1_pos": Column(str, required=False, nullable=True),
        "opp_on_2": Column(str, required=False, nullable=True),
        "opp_on_2_id": Column(str, required=False, nullable=True),
        "opp_on_2_pos": Column(str, required=False, nullable=True),
        "opp_on_3": Column(str, required=False, nullable=True),
        "opp_on_3_id": Column(str, required=False, nullable=True),
        "opp_on_3_pos": Column(str, required=False, nullable=True),
        "opp_on_4": Column(str, required=False, nullable=True),
        "opp_on_4_id": Column(str, required=False, nullable=True),
        "opp_on_4_pos": Column(str, required=False, nullable=True),
        "opp_on_5": Column(str, required=False, nullable=True),
        "opp_on_5_id": Column(str, required=False, nullable=True),
        "opp_on_5_pos": Column(str, required=False, nullable=True),
        "opp_on_6": Column(str, required=False, nullable=True),
        "opp_on_6_id": Column(str, required=False, nullable=True),
        "opp_on_6_pos": Column(str, required=False, nullable=True),
        "opp_on_7": Column(str, required=False, nullable=True),
        "opp_on_7_id": Column(str, required=False, nullable=True),
        "opp_on_7_pos": Column(str, required=False, nullable=True),
        "num_on": Column(float, nullable=True),
        "num_off": Column(float, nullable=True),
        "players_on": Column(str, nullable=True),
        "players_on_id": Column(str, nullable=True),
        "players_on_pos": Column(str, nullable=True),
        "players_off": Column(str, nullable=True),
        "players_off_id": Column(str, nullable=True),
        "players_off_pos": Column(str, nullable=True),
        "zone_start": Column(str, nullable=True),
        "goal": Column(int, default=0),
        "hd_goal": Column(int, default=0),
        "goal_adj": Column(float, default=0),
        "pred_goal": Column(float, nullable=True),
        "pred_goal_adj": Column(float, nullable=True),
        "shot": Column(int, default=0),
        "hd_shot": Column(int, default=0),
        "shot_adj": Column(float, default=0),
        "miss": Column(int, default=0),
        "hd_miss": Column(int, default=0),
        "fenwick": Column(int, default=0),
        "hd_fenwick": Column(int, default=0),
        "fenwick_adj": Column(float, default=0),
        "corsi": Column(int, default=0),
        "corsi_adj": Column(float, default=0),
        "block": Column(int, default=0),
        "hit": Column(int, default=0),
        "give": Column(int, default=0),
        "take": Column(int, default=0),
        "fac": Column(int, default=0),
        "pen0": Column(int, default=0),
        "pen2": Column(int, default=0),
        "pen4": Column(int, default=0),
        "pen5": Column(int, default=0),
        "pen10": Column(int, default=0),
        "change": Column(int, default=0),
        "stop": Column(int, default=0),
        "ozf": Column(int, default=0),
        "nzf": Column(int, default=0),
        "dzf": Column(int, default=0),
        "ozs": Column(int, default=0),
        "nzs": Column(int, default=0),
        "dzs": Column(int, default=0),
        "otf": Column(int, default=0),
    },
    index=Index(int),
    strict="filter",
    add_missing_columns=True,
    coerce=True,
    ordered=True,
)

StatSchema = DataFrameSchema(
    columns={
        "season": Column(str),
        "session": Column(str),
        "game_id": Column(str, required=False),
        "game_date": Column(str, required=False),
        "player": Column(str),
        "player_id": Column(str),
        "position": Column(str),
        "team": Column(str),
        "opp_team": Column(str, required=False),
        "strength_state": Column(str),
        "period": Column(int, required=False),
        "score_state": Column(str, required=False),
        "forwards": Column(str, required=False),
        "forwards_eh_id": Column(str, required=False),
        "forwards_api_id": Column(str, required=False),
        "defense": Column(str, required=False),
        "defense_eh_id": Column(str, required=False),
        "defense_api_id": Column(str, required=False),
        "own_goalie": Column(str, required=False),
        "own_goalie_eh_id": Column(str, required=False),
        "own_goalie_api_id": Column(str, required=False),
        "opp_forwards": Column(str, required=False),
        "opp_forwards_eh_id": Column(str, required=False),
        "opp_forwards_api_id": Column(str, required=False),
        "opp_defense": Column(str, required=False),
        "opp_defense_eh_id": Column(str, required=False),
        "opp_defense_api_id": Column(str, required=False),
        "opp_goalie": Column(str, required=False),
        "opp_goalie_eh_id": Column(str, required=False),
        "opp_goalie_api_id": Column(str, required=False),
        "toi": Column(float, default=0),
        "g": Column(int, default=0),
        "ihdg": Column(int, default=0),
        "a1": Column(int, default=0),
        "a2": Column(int, default=0),
        "ixg": Column(float, default=0),
        "isf": Column(int, default=0),
        "ihdsf": Column(int, default=0),
        "imsf": Column(int, default=0),
        "ihdm": Column(int, default=0),
        "iff": Column(int, default=0),
        "ihdf": Column(int, default=0),
        "isb": Column(int, default=0),
        "icf": Column(int, default=0),
        "ibs": Column(int, default=0),
        "igive": Column(int, default=0),
        "itake": Column(int, default=0),
        "ihf": Column(int, default=0),
        "iht": Column(int, default=0),
        "ifow": Column(int, default=0),
        "ifol": Column(int, default=0),
        "iozfw": Column(int, default=0),
        "iozfl": Column(int, default=0),
        "inzfw": Column(int, default=0),
        "inzfl": Column(int, default=0),
        "idzfw": Column(int, default=0),
        "idzfl": Column(int, default=0),
        "a1_xg": Column(float, default=0),
        "a2_xg": Column(float, default=0),
        "ipent0": Column(int, default=0),
        "ipent2": Column(int, default=0),
        "ipent4": Column(int, default=0),
        "ipent5": Column(int, default=0),
        "ipent10": Column(int, default=0),
        "ipend0": Column(int, default=0),
        "ipend2": Column(int, default=0),
        "ipend4": Column(int, default=0),
        "ipend5": Column(int, default=0),
        "ipend10": Column(int, default=0),
        "gf": Column(int, default=0),
        "ga": Column(int, default=0),
        "hdgf": Column(int, default=0),
        "hdga": Column(int, default=0),
        "xgf": Column(float, default=0),
        "xga": Column(float, default=0),
        "sf": Column(int, default=0),
        "sa": Column(int, default=0),
        "hdsf": Column(int, default=0),
        "hdsa": Column(int, default=0),
        "ff": Column(int, default=0),
        "fa": Column(int, default=0),
        "hdff": Column(int, default=0),
        "hdfa": Column(int, default=0),
        "cf": Column(int, default=0),
        "ca": Column(int, default=0),
        "bsf": Column(int, default=0),
        "bsa": Column(int, default=0),
        "msf": Column(int, default=0),
        "msa": Column(int, default=0),
        "hdmsf": Column(int, default=0),
        "hdmsa": Column(int, default=0),
        "teammate_block": Column(int, default=0),
        "hf": Column(int, default=0),
        "ht": Column(int, default=0),
        "give": Column(int, default=0),
        "take": Column(int, default=0),
        "ozf": Column(int, default=0),
        "nzf": Column(int, default=0),
        "dzf": Column(int, default=0),
        "fow": Column(int, default=0),
        "fol": Column(int, default=0),
        "ozfw": Column(int, default=0),
        "ozfl": Column(int, default=0),
        "nzfw": Column(int, default=0),
        "nzfl": Column(int, default=0),
        "dzfw": Column(int, default=0),
        "dzfl": Column(int, default=0),
        "pent0": Column(int, default=0),
        "pent2": Column(int, default=0),
        "pent4": Column(int, default=0),
        "pent5": Column(int, default=0),
        "pent10": Column(int, default=0),
        "pend0": Column(int, default=0),
        "pend2": Column(int, default=0),
        "pend4": Column(int, default=0),
        "pend5": Column(int, default=0),
        "pend10": Column(int, default=0),
        "ozs": Column(int, default=0),
        "nzs": Column(int, default=0),
        "dzs": Column(int, default=0),
        "otf": Column(int, default=0),
    },
    strict="filter",
    add_missing_columns=True,
    coerce=True,
    ordered=True,
)

LineSchema = DataFrameSchema(
    columns={
        "season": Column(str),
        "session": Column(str),
        "game_id": Column(str, required=False),
        "game_date": Column(str, required=False),
        "team": Column(str),
        "opp_team": Column(str, required=False),
        "strength_state": Column(str),
        "period": Column(int, required=False),
        "score_state": Column(str, required=False),
        "forwards": Column(str, required=False),
        "forwards_id": Column(str, required=False),
        "defense": Column(str, required=False),
        "defense_id": Column(str, required=False),
        "own_goalie": Column(str, required=False),
        "own_goalie_id": Column(str, required=False),
        "opp_forwards": Column(str, required=False),
        "opp_forwards_id": Column(str, required=False),
        "opp_defense": Column(str, required=False),
        "opp_defense_id": Column(str, required=False),
        "opp_goalie": Column(str, required=False),
        "opp_goalie_id": Column(str, required=False),
        "toi": Column(float, default=0),
        "gf": Column(int, default=0),
        "ga": Column(int, default=0),
        "hdgf": Column(int, default=0),
        "hdga": Column(int, default=0),
        "xgf": Column(float, default=0),
        "xga": Column(float, default=0),
        "sf": Column(int, default=0),
        "sa": Column(int, default=0),
        "hdsf": Column(int, default=0),
        "hdsa": Column(int, default=0),
        "ff": Column(int, default=0),
        "fa": Column(int, default=0),
        "hdff": Column(int, default=0),
        "hdfa": Column(int, default=0),
        "cf": Column(int, default=0),
        "ca": Column(int, default=0),
        "bsf": Column(int, default=0),
        "bsa": Column(int, default=0),
        "msf": Column(int, default=0),
        "msa": Column(int, default=0),
        "hdmsf": Column(int, default=0),
        "hdmsa": Column(int, default=0),
        "teammate_block": Column(int, default=0),
        "hf": Column(int, default=0),
        "ht": Column(int, default=0),
        "give": Column(int, default=0),
        "take": Column(int, default=0),
        "ozf": Column(int, default=0),
        "nzf": Column(int, default=0),
        "dzf": Column(int, default=0),
        "fow": Column(int, default=0),
        "fol": Column(int, default=0),
        "ozfw": Column(int, default=0),
        "ozfl": Column(int, default=0),
        "nzfw": Column(int, default=0),
        "nzfl": Column(int, default=0),
        "dzfw": Column(int, default=0),
        "dzfl": Column(int, default=0),
        "pent0": Column(int, default=0),
        "pent2": Column(int, default=0),
        "pent4": Column(int, default=0),
        "pent5": Column(int, default=0),
        "pent10": Column(int, default=0),
        "pend0": Column(int, default=0),
        "pend2": Column(int, default=0),
        "pend4": Column(int, default=0),
        "pend5": Column(int, default=0),
        "pend10": Column(int, default=0),
    },
    strict="filter",
    add_missing_columns=True,
    coerce=True,
    ordered=True,
)



import pandas as pd
import pytest

from chickenstats.api.api import ChickenStats


@pytest.mark.skip(reason="Not working yet")
class TestChickenStats:
    def test_check_pbp_game_ids(self):
        api_instance = ChickenStats()

        game_ids = api_instance.check_pbp_game_ids(season=["20232024"], disable_progress_bar=True)

        assert isinstance(game_ids, list) is True

    def test_check_check_pbp_play_ids(self):
        api_instance = ChickenStats()

        play_ids = api_instance.check_pbp_play_ids(season=["20232024"], disable_progress_bar=True)

        assert isinstance(play_ids, list) is True

    def test_download_pbp(self):
        api_instance = ChickenStats()

        pbp_data = api_instance.download_pbp(game_id=[2023020001], strength_state=["5v5"], disable_progress_bar=True)

        assert isinstance(pbp_data, pd.DataFrame) is True

    def test_check_stats_game_ids(self):
        api_instance = ChickenStats()

        game_ids = api_instance.check_stats_game_ids(season=["20232024"], disable_progress_bar=True)

        assert isinstance(game_ids, list) is True

    def test_download_game_stats(self):
        api_instance = ChickenStats()

        game_stats = api_instance.download_game_stats(
            game_id=[2023020001], strength_state=["5v5"], disable_progress_bar=True
        )

        assert isinstance(game_stats, pd.DataFrame) is True

    def test_check_team_stats_game_ids(self):
        api_instance = ChickenStats()

        game_ids = api_instance.check_team_stats_game_ids(season=["20232024"], disable_progress_bar=True)

        assert isinstance(game_ids, list) is True

    def test_check_lines_game_ids(self):
        api_instance = ChickenStats()

        game_ids = api_instance.check_lines_game_ids(season=["20232024"], disable_progress_bar=True)

        assert isinstance(game_ids, list) is True



from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from chickenstats.chicken_nhl._helpers import convert_to_list, norm_coords, return_name_html


@pytest.fixture(scope="package")
def test_pbp():
    filepath = Path("./tests/tests_chicken_nhl/data/test_pbp.csv")

    test_pbp = pd.read_csv(filepath, low_memory=False)

    return test_pbp


def test_norm_coords(test_pbp, norm_column="event_team", norm_value="NSH"):
    data = norm_coords(data=test_pbp, norm_column=norm_column, norm_value=norm_value)

    if "norm_coords_x" in data.columns and "norm_coords_y" in data.columns:
        assert True

    else:
        raise AssertionError()


@pytest.mark.parametrize(
    "test_list",
    [2023020001, 2023020001.0, "2023020001", (2023020001, 2023020002), pd.Series([2023020001]), np.array([2023020001])],
)
def test_convert_to_list(test_list):
    test_list = convert_to_list(test_list, "GAME ID")

    assert isinstance(test_list, list) is True


def test_convert_to_list_fail():
    with pytest.raises(Exception):
        convert_to_list({"test_key": "test_value"}, "TEST")


def test_return_name_html():
    name = return_name_html("GOALIE - PEKKA RINNE")

    assert isinstance(name, str) is True
